from functools import cached_property
import re

from requests.packages.urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
import requests

from .. import PDU, PDUPort, PDUState


class ShellyPDU(PDU):
    @property
    def requests_retry_session(self, retries=3, backoff_factor=1,
                               status_forcelist=[], session=None):  # pragma: nocover
        session = session or requests.Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        return session

    def url(self, path):
        hostname = self.config['hostname']
        return f"http://{hostname}{path}"

    def get(self, path):
        r = self.requests_retry_session.get(self.url(path), timeout=1)
        r.raise_for_status()
        return r.json()

    @property
    def gen(self):
        if gen := self.raw_dev.get('gen'):  # Gen 2+ hardware
            return gen
        else:
            return 1

    @cached_property
    def num_ports(self):
        # NOTE: We really want to have a stable amount of ports, even in the presence
        # of network issues, so let's depend on the one thing we have: `/shelly`

        if self.gen == 1:
            return self.raw_dev.get('num_outputs')
        elif app := self.raw_dev.get('app'):
            if app in ['PlusPlugS']:
                return 1
            elif m := re.match(r'^(Pro|Plus)(\d)(PM)?$', app):
                return int(m.groups()[1])

        # Assume the worst: no controllable ports
        return 0

    def __init__(self, name, config, reserved_port_ids=[]):
        super().__init__(name, config, reserved_port_ids)

        self.raw_dev = self.get('/shelly')

        if self.gen not in [1, 2] or self.num_ports == 0:
            t = self.raw_dev.get('type')
            m = self.raw_dev.get('model')
            a = self.raw_dev.get('app')
            raise ValueError(f"Unknown Shelly device: gen={self.gen}, type={t}, model={m}, app={a}")

        self._ports = [PDUPort(self, str(oid)) for oid in range(self.num_ports)]

    @property
    def ports(self):
        # Update the ports, in case their name changed
        for port in self._ports:
            if self.gen == 1:
                raw_port = self.get(f'/settings/relay/{port.port_id}')
                port.label = raw_port.get('name')
            else:
                raw_port = self.get(f'/rpc/Switch.GetConfig?id={port.port_id}')
                port.label = raw_port.get('name')

        return self._ports

    def __ison_to_PDUState(self, ison):
        return PDUState.ON if ison else PDUState.OFF

    def set_port_state(self, port_id, state):
        if self.gen == 1:
            raw_port = self.get(f'/relay/{port_id}?turn={state.name.lower()}')
            return self.__ison_to_PDUState(raw_port.get('ison')) == state
        else:
            ison = str(state == PDUState.ON).lower()
            self.get(f'/rpc/Switch.Set?id={port_id}&on={ison}')
            return True

    def get_port_state(self, port_id):
        if self.gen == 1:
            raw_port = self.get(f'/relay/{port_id}')
            return self.__ison_to_PDUState(raw_port.get('ison'))
        else:
            raw_port = self.get(f'/rpc/Switch.GetStatus?id={port_id}')
            return self.__ison_to_PDUState(raw_port.get('output'))



#
# The MIT License (MIT)
#
# Copyright (c) 2015 PsychoMario (imported from PyPXE)
# Copyright (c) 2023 Martin Roukala
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from functools import cached_property
from enum import Enum, auto
import fcntl
import traceback
import socket
import struct
import logging
import re
from collections import defaultdict
from time import time


class MacAddress:
    def __init__(self, macaddr):
        if isinstance(macaddr, bytes):
            self.as_bytes = macaddr
        elif isinstance(macaddr, str):
            self.as_bytes = self.from_human(macaddr)
        else:
            raise ValueError(f"Unknown type {type(macaddr)}")

        # Now that we decoded the mac address, let's generate a uniform mac address
        self.as_str = self.to_human(self.as_bytes)

    @classmethod
    def to_human(cls, raw_mac):
        '''
            This method converts the MAC Address from binary to
            human-readable format for logging.
        '''
        return ':'.join(map(lambda x: hex(x)[2:].zfill(2), struct.unpack('BBBBBB', raw_mac)))

    @classmethod
    def from_human(cls, mac_addr):
        m = re.match("[0-9a-f]{2}([-:]?)[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$",
                     mac_addr.lower())
        if not m:
            raise ValueError(f"{mac_addr} is not a valid mac address")
        else:
            delim = m.groups()[0]
            if delim:
                mac_bytes_list = mac_addr.split(delim)
            else:
                mac_bytes_list = [mac_addr[i:i+2] for i in range(0, len(mac_addr), 2)]

            return bytes.fromhex(" ".join(mac_bytes_list))

    def __str__(self):
        return self.as_str

    def __hash__(self):
        return hash(self.as_bytes)

    def __eq__(self, other):
        return self.as_bytes == other.as_bytes


class DhcpRequestType53(Enum):
    DISCOVER = 1
    REQUEST = 3

    @property
    def response_value(self):
        if self.name == "DISCOVER":
            return 2  # OFFER
        elif self.name == "REQUEST":
            return 5  # ACK


class CPUArch(Enum):
    X86 = auto()
    X86_64 = auto()
    ARM32 = auto()
    ARM64 = auto()
    RISCV32 = auto()
    RISCV64 = auto()

    @property
    def to_ipxe_buildarch(self):
        if self == CPUArch.X86:
            return "i386"
        elif self in [CPUArch.X86_64, CPUArch.ARM32, CPUArch.ARM64]:
            return self.name.lower()
        else:
            return None

    def __str__(self):
        return self.name


class Firmware(Enum):
    BIOS = auto()
    UEFI = auto()
    UBOOT = auto()
    RPI = auto()

    @property
    def to_ipxe_platform(self):
        if self == Firmware.UEFI:
            return "efi"
        elif self == Firmware.BIOS:
            return "pcbios"
        else:
            return None

    def __str__(self):
        return self.name


class BootProtocol(Enum):
    TFTP = auto()
    HTTP = auto()

    def __str__(self):
        return self.name


class ClientRequest:
    def __init__(self, raw_request):
        self.raw_request = raw_request

        # Parse the architecture field (Opt93)
        self.architecture = self.firmware = self.protocol = None
        if decoded_arch := self.client_system_architecture:
            self.architecture, self.firmware, self.protocol = decoded_arch

    @cached_property
    def client_system_architecture_id(self):
        if opt93 := self.raw_dhcp_options.get(93):
            assert len(opt93) == 2
            [arch] = struct.unpack("!H", self.raw_dhcp_options[93])
            return arch

    @cached_property
    def client_system_architecture(self):
        if self.client_system_architecture_id is not None:
            # DHCPv4 and v6 are supposed to have the same archicture mapping,
            # but they ended up disagreeing on whether EFI x86-64 was ID 7 or 9,
            # and it doesn't seem like it got clarified[3]. Given how unlikely
            # bytecode EFI implementations are, let's just pretend both are EFI
            # x86-64 :)
            # [1]: https://www.rfc-editor.org/rfc/rfc4578#section-2.1
            # [2]: https://www.iana.org/assignments/dhcpv6-parameters/dhcpv6-parameters.xhtml#processor-architecture
            # [3]: https://www.syslinux.org/archives/2014-October/022684.html
            known_architectures = {
                # X86 (TFTP)
                0x00: (CPUArch.X86, Firmware.BIOS, BootProtocol.TFTP),
                0x06: (CPUArch.X86, Firmware.UEFI, BootProtocol.TFTP),
                0x07: (CPUArch.X86_64, Firmware.UEFI, BootProtocol.TFTP),
                0x09: (CPUArch.X86_64, Firmware.UEFI, BootProtocol.TFTP),

                # ARM 32-64 (TFTP)
                0x0a: (CPUArch.ARM32, Firmware.UEFI, BootProtocol.TFTP),
                0x0b: (CPUArch.ARM64, Firmware.UEFI, BootProtocol.TFTP),

                # HTTP Boot
                0x0f: (CPUArch.X86, Firmware.UEFI, BootProtocol.HTTP),
                0x10: (CPUArch.X86_64, Firmware.UEFI, BootProtocol.HTTP),
                0x12: (CPUArch.ARM32, Firmware.UEFI, BootProtocol.HTTP),
                0x13: (CPUArch.ARM64, Firmware.UEFI, BootProtocol.HTTP),
                0x14: (CPUArch.X86, Firmware.BIOS, BootProtocol.HTTP),

                # ARM 32-64 U-boot (TFTP & HTTP)
                0x15: (CPUArch.ARM32, Firmware.UBOOT, BootProtocol.TFTP),
                0x16: (CPUArch.ARM64, Firmware.UBOOT, BootProtocol.TFTP),
                0x17: (CPUArch.ARM32, Firmware.UBOOT, BootProtocol.HTTP),
                0x18: (CPUArch.ARM64, Firmware.UBOOT, BootProtocol.HTTP),

                # RISC-V EFI (TFTP & HTTP)
                0x19: (CPUArch.RISCV32, Firmware.UEFI, BootProtocol.TFTP),
                0x1a: (CPUArch.RISCV32, Firmware.UEFI, BootProtocol.HTTP),
                0x1b: (CPUArch.RISCV64, Firmware.UEFI, BootProtocol.TFTP),
                0x1c: (CPUArch.RISCV64, Firmware.UEFI, BootProtocol.HTTP),

                # Raspberry Pi 4 / CM4 (may be limited to IPv6)
                0x29: (CPUArch.ARM64, Firmware.RPI, BootProtocol.TFTP),
            }

            return known_architectures.get(self.client_system_architecture_id)

    @property
    def requested_fields(self):
        return set(self.raw_dhcp_options.get(55, b''))

    @property
    def is_valid_netboot_request(self):
        ''' Verify that all the mandatory fields are present, as mandated by rfc4578 '''

        if 93 not in self.raw_dhcp_options:
            # We are not given the client architecture ID, we can't know what to send!
            return False
        elif not self.requested_fields.issuperset({67}) and self.firmware != Firmware.UBOOT:
            # We are not asked for the boot target, sounds really fishy
            # U-Boot doesn't include this option, but is happy to process it,
            # so allow it just for U-Boot.
            return False
        else:
            return True

        # The following *are* mandated by the spec, but many implementations don't actually
        # provide them. So let's be defensive, and only require fields that we *need* as
        # DHCP clients will ignore
        #
        # if not self.requested_fields.issuperset({128, 129, 130, 131, 132, 133, 134, 135}):
        #     return False
        # elif 94 not in self.raw_dhcp_options:
        #     return False
        # elif 97 not in self.raw_dhcp_options:
        #     return False

    @cached_property
    def mac_addr(self):
        [client_mac] = struct.unpack('!28x6s', self.raw_request[:34])
        return MacAddress(client_mac)

    @cached_property
    def raw_dhcp_options(self):
        '''Parse a string of TLV-encoded options.'''
        raw = self.raw_request[240:]
        ret = {}
        while raw:
            [tag] = struct.unpack('B', raw[0:1])
            if tag == 0:       # padding
                raw = raw[1:]
                continue
            if tag == 255:     # end marker
                break
            [length] = struct.unpack('B', raw[1:2])
            value = raw[2:2 + length]
            raw = raw[2 + length:]
            ret[tag] = value
        return ret

    @cached_property
    def req_type(self):
        if opt53 := self.raw_dhcp_options.get(53):
            return DhcpRequestType53(ord(opt53))  # see RFC2131, page 10

    def __opt_to_str(self, opt):
        if val := self.raw_dhcp_options.get(opt):
            return val.decode('ascii')
        return None

    @cached_property
    def vendor_class(self):
        return self.__opt_to_str(60)

    @cached_property
    def user_class(self):
        return self.__opt_to_str(77)

    def __str__(self):
        mac_addr = self.mac_addr.as_str

        if self.req_type is None:  # pragma: nocover
            return f"InvalidDHCPRequest<{mac_addr}>"
        req_type = self.req_type.name

        vendor_class = self.vendor_class
        user_class = self.user_class
        str_class = user_class if user_class else vendor_class

        details = ""
        if self.client_system_architecture:
            arch = self.architecture.name
            firmware = self.firmware.name
            protocol = self.protocol.name
            details = f" ({arch}/{firmware}/{protocol})"

        return f"DHCP{req_type}<{mac_addr}/{str_class}{details}>"


class OutOfLeasesError(Exception):
    pass


class DHCPD:
    # Callbacks that should be implemented by the user
    @property
    def static_clients(self):  # pragma: nocover
        """" Return here a dictionary that maps a client MAC address to it
        Example: [{"mac_addr": "00:01:02:03:04:05", "ipaddr": "10.0.0.1"}}]
        """

        return []

    def boot_target(self, client_request):  # pragma: nocover
        raise NotImplementedError

    # Implementation details

    '''
        This class implements a DHCP Server, limited to PXE options.
        Implemented from RFC2131, RFC2132,
        https://en.wikipedia.org/wiki/Dynamic_Host_Configuration_Protocol,
        and http://www.pix.net/software/pxeboot/archive/pxespec.pdf.
    '''
    def __init__(self, interface, **kwargs):
        self.interface = interface
        self.broadcast = '255.255.255.255'

        self.router = kwargs.get('router', self.ip)
        self.dns_servers = kwargs.get('dns_servers', ["9.9.9.9"])
        self.file_server = kwargs.get('file_server', self.ip)   # TFTP / HTTP Server
        self.lease_time = kwargs.get('lease_time', 86400)

        self.logger = kwargs.get('logger', None)

        # setup logger
        if self.logger is None:
            self.logger = logging.getLogger('DHCP')
            self.logger.propagate = False
            handler = logging.StreamHandler()
            formatter = logging.Formatter('[%(levelname)s] %(message)s')
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)
            self.logger.setLevel(logging.INFO)

        # In-memory storage of leases, using the MAC address as a key
        self.leases = defaultdict(lambda: {'ip': '', 'expire': 0})

    @classmethod
    def __iface_query_param(cls, iface, param):
        # Implementation from:
        # https://code.activestate.com/recipes/439094-get-the-ip-address-associated-with-a-network-inter
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            try:
                return socket.inet_ntop(socket.AF_INET,
                                        fcntl.ioctl(s.fileno(), param,
                                                    struct.pack('256s',
                                                                iface.encode('utf8'))
                                                    )[20:24])
            except OSError:
                # Iface doesn't exist, or no IP assigned
                raise ValueError(f"The interface {iface} has no IP assigned") from None

    @property
    def ip(self):
        """" Returns the ipv4 address, or None if the iface is unavailable or has
        no IP assigned."""
        return self.__iface_query_param(self.interface, 0x8915)  # SIOCGIFADDR

    @property
    def netmask(self):
        """" Returns the ipv4 netmask, or None if the iface is unavailable or has
        no IP assigned."""
        return self.__iface_query_param(self.interface, 0x891b)  # SIOCGIFNETMASK

    def find_first_available_ip(self):
        '''
            This method returns the next unleased IP from range;
            also does lease expiry by overwrite.
        '''

        # If we use ints, we don't have to deal with octet overflow
        # or nested loops (up to 3 with 10/8); convert both to 32-bit integers
        # e.g '192.168.1.1' <-> 3232235777
        def encode(x):
            return struct.unpack('!I', socket.inet_aton(x))[0]

        def decode(x):
            return socket.inet_ntoa(struct.pack('!I', x))

        # Use the current network settings to identify the list of valid IPs
        nmask = encode(self.netmask)
        network = encode(self.ip) & nmask
        from_host = network + 1
        to_host = (network | ((~ nmask) & 0xffffffff)) - 1

        # Create the list of used IP addresses
        reserved = [self.leases[i]['ip'] for i in self.leases if self.leases[i]['expire'] > time()]
        reserved.extend([c["ipaddr"] for c in self.static_clients if "ipaddr" in c])
        reserved.append(self.ip)                # pull out our own IP
        reserved = list(map(encode, reserved))  # convert to 32-bit int

        # loop through, make sure not already reserved and not in form X.Y.Z.0
        for offset in range(to_host - from_host):
            cur_addr = from_host + offset
            if cur_addr % 256 and cur_addr not in reserved:
                return decode(cur_addr)

        # Couldn't find any available IP, bail out!
        raise OutOfLeasesError('Ran out of IP addresses to lease!')  # pragma: nocover

    def get_or_assign_ip_for_client(self, client_mac):
        # Use the defined static IP, if available
        for client in self.static_clients:
            if client.get("mac_addr") == client_mac:
                return client.get("ipaddr"), "static client"

        # Re-use an existing lease, if we do have one
        if self.leases[client_mac]['ip'] and self.leases[client_mac]['expire'] > time():
            return self.leases[client_mac]['ip'], "rebinding lease"

        # Find the first available IP
        return self.find_first_available_ip(), "new client"

    @classmethod
    def tlv_encode(cls, tag, value):
        '''Encode a TLV option.'''
        if isinstance(value, str):
            value = value.encode('ascii')
        value = bytes(value)
        return struct.pack('BB', tag, len(value)) + value

    def craft_response(self, client_request, offer, boot_target):
        '''This method crafts the full response to the request'''

        # The header
        xid, _, _, _, chaddr = struct.unpack('!4x4s2x2s4x4s4x4s16s', client_request.raw_request[:44])
        response = struct.pack('!BBBB4s',
                               2,                      # Boot reply
                               1,                      # Hardware type: Ethernet
                               6,                      # MAC Address length
                               0,                      # Hops
                               xid)                    # Transaction ID (copied from the request)

        response += struct.pack('!HHI', 0, 0x8000, 0)  # BOOTP flags (Broadcast = 1)

        response += socket.inet_aton(offer)            # Client IP address

        response += socket.inet_aton(self.ip)          # Server IP address
        response += socket.inet_aton('0.0.0.0')        # Relay Agent IP (N/A)
        response += chaddr                             # Client MAC address

        # BOOTP legacy pad
        response += b'\x00' * 64                       # server name
        response += b'\x00' * 128                      # filename
        response += struct.pack('!I', 0x63825363)      # magic cookie

        # The DHCP options
        response += self.tlv_encode(53, struct.pack('!B',
                                                    client_request.req_type.response_value))  # message type
        response += self.tlv_encode(54, socket.inet_aton(self.ip))                            # DHCP Server

        # IP/network configuration
        response += self.tlv_encode(1, socket.inet_aton(self.netmask))                        # subnet mask
        response += self.tlv_encode(3, socket.inet_aton(self.router))                         # router
        dns_servers = b''.join([socket.inet_aton(i) for i in self.dns_servers])
        response += self.tlv_encode(6, dns_servers)                                           # DNS Servers
        response += self.tlv_encode(51, struct.pack('!I', self.lease_time))                   # lease time

        # If we have a boot target / filename, let's send the needed parameters
        if boot_target:
            response += self.tlv_encode(66, self.file_server)                                 # TFTP/HTTP Server
            response += self.tlv_encode(67, boot_target.encode('ascii') + b'\x00')

        response += b'\xff'
        return response

    # TODO: Add tests for this function
    def listen(self, sock=None):  # pragma: nocover
        '''Main listen loop.'''
        def main_loop(sock):
            while True:
                self.current_request = None
                try:
                    raw_request, _ = sock.recvfrom(1024)

                    # Parse the request
                    self.current_request = ClientRequest(raw_request=raw_request)
                    self.logger.info(f"Received the request {str(self.current_request)}")

                    # Find or assign an IP for the client
                    offer, offer_reason = self.get_or_assign_ip_for_client(self.current_request.mac_addr)
                    self.leases[self.current_request.mac_addr]['ip'] = offer
                    self.leases[self.current_request.mac_addr]['expire'] = time() + self.lease_time

                    # Get the expected boot target
                    boot_target = None
                    if self.current_request.is_valid_netboot_request:
                        boot_target = self.boot_target(self.current_request)

                    self.logger.info(f"Offering IP {offer} ({offer_reason}), boot target {boot_target}")

                    # Generate the response
                    if response := self.craft_response(self.current_request, offer, boot_target):
                        sock.sendto(response, (self.broadcast, 68))
                except Exception:
                    self.logger.error(f"Exception caught:\n{traceback.format_exc()}")

        # Use the provided socket if specified, or create our own
        if sock:
            main_loop(sock)
        else:
            with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:  # IPv4 UDP socket
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)  # Allow quick rebinding after restart
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)  # Allow sending packets to a broadcast addr
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_BINDTODEVICE,  # Only listen to msgs from the wanted iface
                                self.interface.encode())
                sock.bind(('', 67))                                         # Bind to port 67

                main_loop(sock)



from unittest.mock import MagicMock, patch, PropertyMock
import pytest
import time

from server.pdu import PDUState
from server.pdu.drivers.snmp import (
    retry_on_known_errors,
    SnmpPDU,
    Session
)


@pytest.fixture(autouse=True)
def reset_easysnmp_mock(monkeypatch):
    import server.pdu.drivers.snmp as snmp

    global Session, time_sleep
    m1, m2 = MagicMock(), MagicMock()
    # REVIEW: I wonder if there's a clever way of covering the
    # difference in import locations between here and snmp.py
    monkeypatch.setattr(snmp, "Session", m1)
    monkeypatch.setattr(time, "sleep", m2)
    Session = m1
    time_sleep = m2


@patch("random.random", return_value=0.42)
def test_driver_BaseSnmpPDU_retry_on_known_errors__known_error(random_mock):
    global retriable_error_call_count
    retriable_error_call_count = 0

    @retry_on_known_errors
    def retriable_error():
        global retriable_error_call_count

        assert time_sleep.call_count == retriable_error_call_count

        retriable_error_call_count += 1
        raise SystemError("<built-in function set> returned NULL without setting an error")

    with pytest.raises(ValueError):
        retriable_error()

    time_sleep.assert_called_with(1.42)
    assert time_sleep.call_count == retriable_error_call_count
    assert retriable_error_call_count == 3


class MockSnmpPDU(SnmpPDU):
    system_id = '1234.1.2.3.4'
    outlet_labels = '4.5.4.5'
    outlet_status = '4.5.4.6'

    state_mapping = {
        PDUState.ON: 2,
        PDUState.OFF: 3,
        PDUState.REBOOT: 4,
    }

    state_transition_delay_seconds = 5


def test_driver_SnmpPDU_eq():
    params = {
        "hostname": "hostname",
        "community": "community"
    }

    assert MockSnmpPDU("name", params) == MockSnmpPDU("name", params)
    for param in params:
        n_params = dict(params)
        n_params[param] = "modified"
        assert MockSnmpPDU("name", params) != MockSnmpPDU("name", n_params)


def test_driver_SnmpPDU_listing_ports():
    walk_mock = Session.return_value.walk
    walk_mock.return_value = [MagicMock(value="P1"), MagicMock(value="P2")]

    pdu = MockSnmpPDU("MyPDU", {"hostname": "127.0.0.1"})
    walk_mock.assert_called_with(pdu.outlet_labels_oid)
    ports = pdu.ports

    # Check that the labels are stored, and the port IDs are 1-indexed
    for i in range(0, 2):
        assert ports[i].port_id == i+1
        assert ports[i].label == f"P{i+1}"


def test_driver_SnmpPDU_listing_ports__walk_failure():
    err_msg = "<built-in function walk> returned NULL without setting an error"
    Session.return_value.walk.side_effect = SystemError(err_msg)

    with pytest.raises(ValueError):
        MockSnmpPDU("MyPDU", {"hostname": "127.0.0.1"})


def test_driver_BaseSnmpPDU_port_label_mapping():
    walk_mock = Session.return_value.walk
    set_mock = Session.return_value.set
    walk_mock.return_value = [
        MagicMock(value="P1"),
        MagicMock(value="P2")
    ]

    pdu = MockSnmpPDU("MyPDU", {"hostname": "127.0.0.1"})

    set_mock.return_value = True
    assert pdu.set_port_state("P1", PDUState.REBOOT) is True
    set_mock.assert_called_with(pdu.outlet_ctrl_oid(1), pdu.state_mapping[PDUState.REBOOT], 'i')
    assert pdu.set_port_state("P2", PDUState.REBOOT) is True
    set_mock.assert_called_with(pdu.outlet_ctrl_oid(2), pdu.state_mapping[PDUState.REBOOT], 'i')
    with pytest.raises(ValueError):
        pdu.set_port_state("flubberbubber", PDUState.OFF)


def test_driver_BaseSnmpPDU_get_port():
    get_mock = Session.return_value.get

    pdu = MockSnmpPDU("MyPDU", {"hostname": "127.0.0.1"})
    get_mock.return_value.value = pdu.state_mapping[PDUState.REBOOT]
    get_mock.assert_called_with(pdu.outlet_status_oid(1))
    pdu_state = pdu.get_port_state(2)
    assert pdu_state == PDUState.REBOOT
    get_mock.assert_called_with(pdu.outlet_status_oid(2))

    get_mock.side_effect = SystemError("<built-in function get> returned NULL without setting an error")
    with pytest.raises(ValueError):
        pdu.get_port_state(2)


def test_driver_BaseSnmpPDU_set_port():
    set_mock = Session.return_value.set

    pdu = MockSnmpPDU("MyPDU", {"hostname": "127.0.0.1"})
    type(set_mock).value = PropertyMock(return_value=pdu.state_mapping[PDUState.REBOOT])
    set_mock.return_value = True
    set_mock.assert_not_called()
    assert pdu.set_port_state(2, PDUState.REBOOT) is True
    set_mock.assert_called_with(pdu.outlet_ctrl_oid(2), pdu.state_mapping[PDUState.REBOOT], 'i')

    set_mock.side_effect = SystemError("<built-in function set> returned NULL without setting an error")
    with pytest.raises(ValueError):
        pdu.set_port_state(2, PDUState.REBOOT)


def test_driver_BaseSnmpPDU_action_translation():
    pdu = MockSnmpPDU("MyPDU", {"hostname": "127.0.0.1"})

    # Check the state -> SNMP value translation
    for action in PDUState.valid_actions():
        assert pdu.inverse_state_mapping[pdu.state_mapping[action]] == action

    with pytest.raises(KeyError):
        pdu.state_mapping[PDUState.UNKNOWN]



#
# The MIT License (MIT)
#
# Copyright (c) 2015 PsychoMario (imported from PyPXE)
# Copyright (c) 2023 Martin Roukala
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import socket
import struct
import os
import select
import time
import logging
import math


class ParentSocket(socket.socket):
    '''Subclassed socket.socket to enable a link-back to the client object.'''
    parent = None


class PathTraversalException(Exception):
    pass


def normalize_path(base, filename):
    '''
        Join and normalize a base path and filename.

       `base` may be relative, in which case it's converted to absolute.

        Args:
            base (str): Base path
            filename (str): Filename (optionally including path relative to
                            base)

        Returns:
            str: The joined and normalized path

        Raises:
            PathTraversalException: if an attempt to escape the base path is
                                    detected
    '''
    abs_path = os.path.abspath(base)
    joined = os.path.join(abs_path, filename)
    normalized = os.path.normpath(joined)
    if normalized.startswith(os.path.join(abs_path, '')):
        return normalized
    raise PathTraversalException('Path Traversal detected')


class TFTPDClient:
    '''Client instance for TFTPD.'''
    def __init__(self, mainsock, parent):
        self.message, self.address = mainsock.recvfrom(1024)
        self.logger = logging.getLogger(f"{parent.logger.name}.Client.{self.address}")
        self.parent = parent
        self.logger.debug('Receiving request...')
        self.retries = self.parent.default_retries
        self.block = 1
        self.blksize = 512
        self.sent_time = float('inf')
        self.dead = False
        self.fh = None
        self.filename = ''
        self.wrap = 0
        self.arm_wrap = False
        self.handle()  # message from the main socket

    def ready(self):
        '''Called when there is something to be read on our socket.'''
        self.message = self.sock.recv(1024)
        self.handle()

    def send_block(self):
        '''
            Sends the next block of data, setting the timeout and retry
            variables accordingly.
        '''

        # We never started a transfer, abort!
        if self.block == 0:
            self.complete()
            return

        data = None
        try:
            self.fh.seek(self.blksize * (self.block - 1))
            data = self.fh.read(self.blksize)
        except Exception:
            self.logger.error('Error while reading block {0}'.format(self.block))
            self.dead = True
            return
        # opcode 3 == DATA, wraparound block number
        response = struct.pack('!HH', 3, self.block % 65536)
        response += data
        self.sock.sendto(response, self.address)
        self.logger.debug('Sending block {0}/{1}'.format(self.block, self.lastblock))
        self.retries -= 1
        self.sent_time = time.time()

    def no_ack(self):
        '''Determines if we timed out waiting for an ACK from the client.'''
        if self.sent_time + self.parent.timeout < time.time():
            return True
        return False

    def no_retries(self):
        '''Determines if the client ran out of retry attempts.'''
        if not self.retries:
            return True
        return False

    def valid_mode(self):
        '''Determines if the file read mode octet; if not, send an error.'''
        mode = self.message.split(b'\x00')[1]
        if mode == b'octet':
            return True
        self.send_error(5, 'Mode {0} not supported'.format(mode))
        return False

    def load_file(self, filename):
        '''
            Determines if the file exists under the netboot_directory,
            and if it is a file; if not, send an error.
        '''
        try:
            filename = normalize_path(self.parent.netboot_directory, filename)
        except PathTraversalException:
            return False

        # check that the file is really a file, not a rogue FIFO, socket or
        # a device node lurking inside TFTP dir
        if not os.path.lexists(filename) or not os.path.isfile(filename):
            return False

        try:
            self.fh = open(filename, 'rb')
        except IOError:
            return False

        self.filename = filename

        self.fh.seek(0, os.SEEK_END)
        self.filesize = self.fh.tell()
        self.fh.seek(0, os.SEEK_SET)

        self.logger.info('File {0} ({1} bytes) requested'.format(self.filename, self.filesize))

        return True

    def parse_options(self):
        '''
            Extracts the options sent from a client; if any, calculates the last
            block based on the filesize and blocksize.
        '''
        options = self.message.split(b'\x00')[2: -1]
        options = dict(zip((i.decode('ascii') for i in options[0::2]), map(int, options[1::2])))
        self.changed_blksize = 'blksize' in options
        if self.changed_blksize:
            self.blksize = options['blksize']
        self.lastblock = math.ceil(self.filesize / float(self.blksize))
        self.tsize = True if 'tsize' in options else False
        if self.filesize > (2 ** 16) * self.blksize:
            self.logger.warning('Request too big, attempting transfer anyway.')
            self.logger.debug('Details: Filesize {0} is too big for blksize {1}.'.format(self.filesize, self.blksize))
        if len(options):
            # we need to know later if we actually had any options
            self.block = 0
            return True
        else:
            return False

    def reply_options(self):
        '''Acknowledges any options received.'''
        # only called if options, so send them all
        response = struct.pack("!H", 6)
        if self.changed_blksize:
            response += b'blksize' + b'\x00'
            response += str(self.blksize).encode('ascii') + b'\x00'
        if self.tsize:
            response += b'tsize' + b'\x00'
            response += str(self.filesize).encode('ascii') + b'\x00'

        self.sent_time = time.time()
        self.sock.sendto(response, self.address)

    def new_request(self):
        '''
            When receiving a read request from the parent socket, open our
            own socket and check the read request; if we don't have any options,
            send the first block.
        '''
        self.sock = ParentSocket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind((self.parent.ip, 0))
        # used by select() to find ready clients
        self.sock.parent = self
        if not self.valid_mode():
            # some clients just ACK the error (wrong code?)
            # so forcefully shutdown
            self.complete()
            return

        filename = self.message.split(b'\x00')[0].decode('ascii').lstrip('/')

        self.logger.info(f"Client requested {filename}")

        if not self.load_file(filename):
            self.send_error(1, 'File Not Found', filename=filename)
            self.complete()
            return

        if not self.parse_options():
            # no options received so start transfer
            if self.block == 1:
                self.send_block()
            return
        self.reply_options()  # we received some options so ACK those first

    def send_error(self, code=1, message='File Not Found', filename=''):
        '''
            Sends an error code and string to a client. See RFC1350, page 10 for
            details.

            Value   Meaning
            =====   =======
            0       Not defined, see error message (if any).
            1       File not found.
            2       Access violation.
            3       Disk full or allocation exceeded.
            4       Illegal TFTP operation.
            5       Unknown transfer ID.
            6       File already exists.
            7       No such user.
        '''
        response = struct.pack('!H', 5)     # error opcode
        response += struct.pack('!H', code)  # error code
        response += message.encode('ascii')
        response += b'\x00'
        self.sock.sendto(response, self.address)
        self.logger.info('Sending {0}: {1} {2}'.format(code, message, filename))

    def complete(self):
        '''
            Closes a file and socket after sending it
            and marks ourselves as dead to be cleaned up.
        '''
        try:
            if self.fh and not self.fh.closed:
                self.fh.close()
        except Exception as e:
            self.logger.warning(f"Caught the following exception: {e}")

        try:
            if self.sock and self.sock.fileno() >= 0:
                self.sock.shutdown(socket.SHUT_RDWR)
                self.sock.close()
        except Exception as e:
            self.logger.warning(f"Caught the following exception: {e}")

        self.dead = True

    def handle(self):
        '''Takes the message from the parent socket and act accordingly.'''
        # if addr not in ongoing, call this, else ready()
        [opcode] = struct.unpack('!H', self.message[:2])
        if opcode == 1:
            self.message = self.message[2:]
            self.new_request()
        elif opcode == 4:
            [block] = struct.unpack('!H', self.message[2:4])
            if block == 0 and self.arm_wrap:
                self.wrap += 1
                self.arm_wrap = False
            if block == 32768:
                self.arm_wrap = True
            if block < self.block % 65536:
                self.logger.warning('Ignoring duplicated ACK received for block {0}'.format(self.block))
            elif block > self.block % 65536:
                self.logger.warning('Ignoring out of sequence ACK received for block {0}'.format(self.block))
            elif block + self.wrap * 65536 == self.lastblock:
                if self.filesize % self.blksize == 0:
                    self.block = block + 1
                    self.send_block()
                self.logger.info('Completed sending {0}'.format(self.filename))
                self.complete()
            else:
                self.block = block + 1
                self.retries = self.parent.default_retries
                self.send_block()
        elif opcode == 2:
            # write request
            self.sock = ParentSocket(socket.AF_INET, socket.SOCK_DGRAM)
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.sock.bind((self.ip, 0))
            # used by select() to find ready clients
            self.sock.parent = self
            # send error
            self.send_error(4, 'Write support not implemented')
            self.dead = True


class TFTPD:
    '''
        This class implements a read-only TFTP server
        implemented from RFC1350 and RFC2348
    '''
    def __init__(self, interface, client=TFTPDClient, **server_settings):
        self.interface = interface
        self.client = client

        self.ip = server_settings.get('ip', '0.0.0.0')
        self.port = int(server_settings.get('port', 69))
        self.netboot_directory = server_settings.get('netboot_directory', '.')
        self.logger = server_settings.get('logger', None)
        self.default_retries = server_settings.get('default_retries', 10)
        self.timeout = server_settings.get('timeout', 1)

        # setup logger
        if self.logger is None:
            self.logger = logging.getLogger('TFTP')
            self.logger.propagate = False
            handler = logging.StreamHandler()
            formatter = logging.Formatter('%(asctime)s [%(levelname)s] %(message)s')
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)
            self.logger.setLevel(logging.INFO)

        self.ongoing = []

    def listen(self, sock=None):
        '''This method listens for incoming requests.'''
        def main_loop(server_sock):
            while True:
                # remove complete clients to select doesn't fail
                for client in self.ongoing:
                    if client.dead:
                        client.complete()
                        self.ongoing.remove(client)

                listen_list = [server_sock] + [client.sock for client in self.ongoing if not client.dead]
                rlist, _, _ = select.select(listen_list, [], [], 1)
                for sock in rlist:
                    if sock == server_sock:
                        # main socket, so new client
                        self.ongoing.append(self.client(sock, self))
                    else:
                        # client socket, so tell the client object it's ready
                        sock.parent.ready()
                # if we haven't received an ACK in timeout time, retry
                [client.send_block() for client in self.ongoing if client.no_ack()]
                # if we have run out of retries, kill the client
                for client in self.ongoing:
                    if client.no_retries():
                        client.logger.info('Timeout while sending {0}'.format(client.filename))
                        client.complete()

        # Use the provided socket if specified, or create our own
        if sock:
            main_loop(sock)
        else:
            with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:      # IPv4 UDP socket
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)      # Allow quick rebinding after restart
                if self.interface:
                    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BINDTODEVICE,  # Only listen to the wanted iface
                                    self.interface.encode())
                sock.bind((self.ip, self.port))
                main_loop(sock)



#!/usr/bin/env python3

from dataclasses import asdict, fields
from datetime import datetime
from functools import cached_property
from http.server import ThreadingHTTPServer, BaseHTTPRequestHandler
from threading import Thread, Event
from collections import namedtuple
from urllib.parse import urlsplit, urlparse
from enum import IntEnum

from cachecontrol import CacheControl
from cachecontrol.caches.file_cache import FileCache
from pydantic import field_validator
from pydantic.dataclasses import dataclass

from .android.fastbootd import FastbootDevice
from .dut import JobRequest, DUTState, lock_fd, PduPortStats
from .mars import MarsDB
from .message import LogLevel, JobIOMessage, ControlMessage, SessionEndMessage, Message, MessageType
from .pdu import PDUState, PDU
from .message import JobStatus
from .job import Job
from .logger import logger
from .minioclient import MinioClient, MinIOPolicyStatement, generate_policy
from . import config
from .boots import BootConfig
from .boots_db import BootsDB

import subprocess
import traceback
import threading
import requests
import tempfile
import logging
import secrets
import random
import select
import string
import shutil
import socket
import struct
import fcntl
import flask
import json
import time
import yaml
import sys
import os


# Constants
CONSOLE_DRAINING_DELAY = 1


def str_to_int(string, default):
    try:
        return int(string)
    except Exception:
        return default


class JobConsoleState(IntEnum):
    CREATED = 0
    ACTIVE = 1
    DUT_DONE = 2
    TEAR_DOWN = 3
    OVER = 4


class JobConsole(Thread):
    def __init__(self, machine_id, client_endpoint,
                 client_version=None, log_level=LogLevel.INFO):
        super().__init__(name='ConsoleThread')

        self.dut_id = machine_id

        self.client_endpoint = client_endpoint
        self.console_patterns = None
        self.client_version = client_version
        self.log_level = log_level

        # Sockets
        if self.client_version:
            logger.info(f"Connecting to the client endpoint {self.client_endpoint}")
            self.client_sock = socket.create_connection((self.client_endpoint.host, self.client_endpoint.port))
        else:
            self.client_sock = None
        self.salad_sock = None

        # Job-long state
        self._state = JobConsoleState.CREATED
        self.start_time = None
        self.line_buffer = bytearray()
        self._user_session_state = dict()

        self.reset_per_boot_state()

    @property
    def machine_is_unfit_for_service(self):
        return self.console_patterns and self.console_patterns.machine_is_unfit_for_service

    @classmethod
    def salad_request(cls, dut_id):
        salad_url = f"{config.SALAD_URL}/api/v1/machine/{dut_id}"
        r = requests.get(salad_url)
        r.raise_for_status()
        return r.json()

    def connect_to_salad(self):
        parsed_url = urlsplit(config.SALAD_URL)
        machine = self.salad_request(self.dut_id)
        port = machine.get("tcp_port")

        return socket.create_connection((parsed_url.hostname, port))

    def reset_per_boot_state(self):
        self.last_activity_from_machine = None
        self.last_activity_from_client = None

        if self.console_patterns:
            self.console_patterns.reset_per_boot_state()
            self.needs_reboot = self.console_patterns.needs_reboot

    def close_salad(self):
        try:
            self.salad_sock.shutdown(socket.SHUT_RDWR)
            self.salad_sock.close()
        except OSError:
            pass

    def close_client(self):
        if self.client_version:
            try:
                self.client_sock.shutdown(socket.SHUT_RDWR)
                self.client_sock.close()
            except OSError:
                pass

    def close(self):
        self.set_state(JobConsoleState.OVER)

    @property
    def state(self):
        if self._state == JobConsoleState.ACTIVE:
            return self._state if self.is_alive() else JobConsoleState.OVER

        return self._state

    def set_state(self, state, **kwargs):
        prev_state = self._state
        if state < prev_state:
            raise ValueError("The state can only move forward")
        elif state == prev_state:
            return
        else:
            self._state = state

        self.log(f"Job console state changed from {prev_state.name} -> {state.name}\n")

        if state == JobConsoleState.ACTIVE:
            self.start_time = datetime.now()

        elif state == JobConsoleState.DUT_DONE:
            # Skip the entire tear-down if we do not have a client
            if not self.client_version:
                self.set_state(JobConsoleState.OVER)

        elif state == JobConsoleState.TEAR_DOWN:
            # Kill the connection to SALAD
            self.close_salad()

            # Notify the client
            if self.client_version:
                if self.client_version == 0:
                    self.log(f"<-- End of the session: {self.console_patterns.job_status} -->\n")
                elif self.client_version == 1:
                    try:
                        status = JobStatus.from_str(self.console_patterns.job_status)
                        SessionEndMessage.create(job_bucket=kwargs.get('job_bucket'),
                                                 status=status).send(self.client_sock)
                    except (ConnectionResetError, BrokenPipeError, OSError):
                        traceback.print_exc()
                try:
                    self.client_sock.shutdown(socket.SHUT_WR)
                except (ConnectionResetError, BrokenPipeError, OSError):
                    pass

        elif state == JobConsoleState.OVER:
            # Make sure the connections to SALAD and the client are killed
            self.close_salad()
            self.close_client()

    def start(self, console_patterns):
        self.console_patterns = console_patterns
        super().start()

    def match_console_patterns(self, buf):
        patterns_matched = set()

        # Process the buffer, line by line
        to_process = self.line_buffer + buf
        cur = 0
        while True:
            idx = to_process.find(b'\n', cur)
            if idx >= 0:
                line = to_process[cur:idx+1]
                logger.info(f"{self.dut_id} -> {bytes(line)}")
                patterns_matched |= self.console_patterns.process_line(line)
                cur = idx + 1
            else:
                break
        self.line_buffer = to_process[cur:]

        # Tell the user what happened
        if len(patterns_matched) > 0:
            self.log(f"Matched the following patterns: {', '.join(patterns_matched)}\n")

        # Check if the state changed
        self.needs_reboot = self.console_patterns.needs_reboot

    def log(self, msg, log_level=LogLevel.INFO):
        # Ignore messages with a log level lower than the minimum set
        if log_level < self.log_level:
            return

        if self.start_time is not None:
            relative_time = (datetime.now() - self.start_time).total_seconds()
        else:
            relative_time = 0.0

        log_msg = f"+{relative_time:.3f}s: {msg}"
        logger.info(log_msg.rstrip("\r\n"))

        if self.client_version:
            try:
                if self.client_version == 0:
                    self.client_sock.send(log_msg.encode())
                elif self.client_version == 1:
                    ControlMessage.create(log_msg, severity=log_level).send(self.client_sock)
            except OSError:
                pass

    def stop(self):
        self.set_state(JobConsoleState.OVER)
        self.join()

    def run(self):
        try:
            self.salad_sock = self.connect_to_salad()
            self.set_state(JobConsoleState.ACTIVE)
        except Exception:
            self.log(f"ERROR: Failed to connect to the SALAD server:\n{traceback.format_exc()}")
            self.close()

        while self.state < JobConsoleState.OVER:
            fds = []
            if self.state < JobConsoleState.TEAR_DOWN:
                fds.extend([self.salad_sock.fileno()])
            if self.client_version:
                fds.extend([self.client_sock.fileno()])

            # Make sure all the FDs are valid, or exit!
            if any([fd < 0 for fd in fds]):
                self.log("Found a negative fd, aborting!")
                self.close()

            rlist, _, _ = select.select(fds, [], [], 1.0)

            for fd in rlist:
                try:
                    if fd == self.salad_sock.fileno():
                        # DUT's stdout/err: Salad -> Client
                        buf = self.salad_sock.recv(8192)
                        if len(buf) == 0:
                            self.set_state(JobConsoleState.DUT_DONE)

                        # Match the console patterns
                        try:
                            self.match_console_patterns(buf)
                        except Exception:
                            self.log(traceback.format_exc())

                        # Update the last console activity if we already had activity,
                        # or when we get the first newline character as serial
                        # consoles may sometimes send unwanted characters at power up
                        if self.last_activity_from_machine is not None or b'\n' in buf:
                            self.last_activity_from_machine = datetime.now()

                        # Forward to the client
                        if self.client_version:
                            if self.client_version == 0:
                                self.client_sock.send(buf)
                            elif self.client_version == 1:
                                JobIOMessage.create(buf).send(self.client_sock)

                        # The message got forwarded, close the session if it ended
                        if self.console_patterns.session_has_ended:
                            self.set_state(JobConsoleState.DUT_DONE)

                    elif self.client_sock and fd == self.client_sock.fileno():
                        # DUT's stdin: Client -> Salad
                        if self.client_version == 0:
                            buf = self.client_sock.recv(8192)
                            if len(buf) == 0:
                                self.close()

                            # Forward to the salad
                            self.salad_sock.send(buf)
                        elif self.client_version == 1:
                            try:
                                msg = Message.next_message(self.client_sock)
                                if msg.msg_type == MessageType.JOB_IO:
                                    self.salad_sock.send(msg.buffer)
                            except EOFError:
                                # Do not warn when we are expecting the client to close its socket
                                if self.state < JobConsoleState.TEAR_DOWN:
                                    self.log(traceback.format_exc())

                                self.log("The client closed its connection")

                                # Clean up everything on our side
                                self.close()

                        self.last_activity_from_client = datetime.now()
                except (ConnectionResetError, BrokenPipeError, OSError):
                    self.log(traceback.format_exc())
                    self.close()
                except Exception:
                    logger.error(traceback.format_exc())


class JobBucket:
    Credentials = namedtuple('Credentials', ['username', 'password', 'policy_name'])

    def __init__(self, minio, bucket_name, initial_state_tarball_file=None,
                 hostname_by_role={}):
        self.minio = minio
        self.name = bucket_name
        self.hostname_by_role = hostname_by_role

        self._credentials = dict()

        if initial_state_tarball_file:
            self.initial_state_tarball_file = tempfile.NamedTemporaryFile("w+b")
            shutil.copyfileobj(initial_state_tarball_file, self.initial_state_tarball_file)
            self.initial_state_tarball_file.seek(0)
        else:
            self.initial_state_tarball_file = None

        # Ensure the bucket doesn't already exist
        if not self.minio.bucket_exists(bucket_name):
            self.minio.make_bucket(bucket_name)
        else:
            raise ValueError("The bucket already exists")

    def remove(self):
        if self.minio.bucket_exists(self.name):
            self.minio.remove_bucket(self.name)

        for credentials in self._credentials.values():
            self.minio.remove_user_policy(credentials.policy_name, credentials.username)
            self.minio.remove_user(credentials.username)
        self._credentials = {}

    def __del__(self):
        try:
            self.remove()
        except Exception:
            traceback.print_exc()

    def credentials(self, role):
        return self._credentials.get(role)

    def create_owner_credentials(self, role, user_name=None, password=None,
                                 groups=None, whitelisted_ips=None):
        if user_name is None:
            user_name = f"{self.name}-{role}"

        if password is None:
            password = secrets.token_hex(16)

        if groups is None:
            groups = []

        if whitelisted_ips is None:
            whitelisted_ips = []

        policy_name = f"policy_{user_name}"

        self.minio.add_user(user_name, password)

        policy_statements = [
            MinIOPolicyStatement(buckets=[self.name], source_ips=whitelisted_ips)
        ]
        if len(whitelisted_ips) > 0:
            restrict_to_whitelisted_ips = MinIOPolicyStatement(allow=False, not_source_ips=whitelisted_ips)
            policy_statements.append(restrict_to_whitelisted_ips)
        policy = json.dumps(generate_policy(policy_statements))
        logger.debug(f"Applying the MinIO policy: {policy}")

        try:
            self.minio.apply_user_policy(policy_name, user_name, policy_statements)
        except Exception as e:
            self.minio.remove_user(user_name)
            raise e from None

        # Add the user to the wanted list of groups
        for group_name in groups:
            self.minio.add_user_to_group(user_name, group_name)

        credentials = self.Credentials(user_name, password, policy_name)
        self._credentials[role] = credentials

        return credentials

    def setup(self):
        if self.initial_state_tarball_file:
            self.minio.extract_archive(self.initial_state_tarball_file, self.name)
            self.initial_state_tarball_file.close()

    def access_url(self, role=None):
        endpoint = urlparse(self.minio.url)

        role_creds = self.credentials(role)
        if role_creds:
            credentials = f"{role_creds[0]}:{role_creds[1]}@"
        else:
            credentials = ""

        hostname = self.hostname_by_role.get(role, endpoint.hostname)
        return f'{endpoint.scheme}://{credentials}{hostname}:{endpoint.port}'

    @classmethod
    def from_job_request(cls, minio, request, machine):
        # Look for the HOST header, to get the hostname used by the client to connect to
        # the executor, so that we can use the same host when telling the client how to
        # download shared folder
        hostname_by_role = {}
        for name, value in request.http_headers.items():
            if name.lower() == "host":
                if len(value) > 0:
                    hostname_by_role["client"] = value.split(":")[0]

        # Convert the job_bucket_initial_state_tarball_file_fd to a file-like object
        if request.job_bucket_initial_state_tarball_file_fd > 0:
            initial_state_tarball_file = os.fdopen(request.job_bucket_initial_state_tarball_file_fd, "rb")
        else:
            initial_state_tarball_file = None

        last_exception = None
        for i in range(5):
            # Make sure the fixed part of the bucket name isn't filling up the whole bucket name (64 chars max)
            base_bucket_name = f"job-{machine.id}-{request.job_id}"[0:56]

            # Append up to 32 characters of entropy within the bucket name limits of minio:
            # Bucket names can consist only of lowercase letters, numbers, dots (.), and hyphens (-)
            # We however do not allow dots, as the following sequence is not allowed: .., .-, and -.
            rnd = ''.join(random.choice(string.ascii_lowercase + string.digits + '-') for i in range(32))

            try:
                bucket_name = MinioClient.create_valid_bucket_name(f"{base_bucket_name}-{rnd}")

                return cls(minio, bucket_name=bucket_name,
                           initial_state_tarball_file=initial_state_tarball_file,
                           hostname_by_role=hostname_by_role)
            except ValueError as e:
                last_exception = e

        raise last_exception from None


class JobHTTPServerRequestHandler(BaseHTTPRequestHandler):
    def log_message(self, fmt, *args):
        self.server.executor.log(fmt % args)

    @property
    def artifacts(self):
        if deployment := self.server.executor.cur_deployment:
            if deployment.storage:
                return deployment.storage.http

        return []

    def get_artifact(self):
        path = self.path.split('?')[0]
        for artifact in self.artifacts:
            if path == artifact.path:
                return artifact

    def handle_request(self, headers_only=False):
        # Let's close the connection after every transaction
        self.close_connection = True

        # Restrict access to the job's artifacts to the DUT that is supposed to access them
        # While restricting access by IP isn't foolproof, it makes it harder for potential
        # attackers to figure out if they are talking to the right HTTP server. To this end,
        # we also do not use the error code 403 so as to make it less clear what is going on
        dut_ip_address = self.server.executor.db_dut.ip_address
        if self.client_address[0] != dut_ip_address:
            self.log_message("WARNING: Got an HTTP query from an unexpected IP address "
                             f"({self.client_address[0]} instead of {dut_ip_address})")
            self.send_error(404)

        if artifact := self.get_artifact():
            # TODO: Add other ways of providing the data
            data = artifact.data

            data_buf = data.encode()

            self.send_response(200)
            self.send_header("Content-Type", "text/plain;charset=UTF-8")  # TODO: Get the mime type from the job?
            self.send_header("Content-Length", f"{len(data_buf)}")
            self.end_headers()

            if not headers_only:
                self.wfile.write(data_buf)
        else:
            self.send_error(404)

    def do_GET(self):
        self.handle_request(headers_only=False)

    def do_HEAD(self):
        self.handle_request(headers_only=True)


class JobHTTPServer(ThreadingHTTPServer):
    @classmethod
    def __iface_query_param(cls, iface, param):
        # Implementation from:
        # https://code.activestate.com/recipes/439094-get-the-ip-address-associated-with-a-network-inter
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            try:
                return socket.inet_ntop(socket.AF_INET,
                                        fcntl.ioctl(s.fileno(), param,
                                                    struct.pack('256s',
                                                                iface.encode('utf8'))
                                                    )[20:24])
            except OSError:
                # Iface doesn't exist, or no IP assigned
                raise ValueError(f"The interface {iface} has no IP assigned") from None

    def __init__(self, executor):
        self.executor = executor

        # Only expose the server to private interface, if set
        host = ""
        if config.PRIVATE_INTERFACE:
            try:
                host = self.__iface_query_param(config.PRIVATE_INTERFACE, 0x8915)  # SIOCGIFADDR
            except Exception:
                self.executor.log(("WARNING: Failed to get the IP address of the private interface:\n"
                                   f"{traceback.format_exc()}"))

        super().__init__((host, 0), JobHTTPServerRequestHandler)
        threading.Thread(target=self.serve_forever, name="HTTP", daemon=True).start()

    @property
    def url(self):
        host, port = self.server_address
        return f"http://{host}:{port}"


class Executor(Thread):
    def __init__(self, mars_db, db_dut, job_request):
        self.mars_db = mars_db
        self.db_dut = db_dut
        self.job_request = job_request

        self.state = DUTState.QUEUED
        self.job_console = None
        self.cur_deployment = None
        self.stop_event = Event()

        # Remote artifacts (typically over HTTPS) are stored in our
        # local minio instance which is exposed over HTTP to the
        # private LAN. This makes such artifacts amenable to PXE
        # booting, for which HTTPS clients are not available.  Less
        # critically, it makes access easier for the boards in our
        # private LAN, for which HTTPS offers no advantage.
        self.remote_url_to_local_cache_mapping = {}

    @cached_property
    def minio(self):
        return MinioClient()

    @cached_property
    def job_bucket(self):
        job_bucket = JobBucket.from_job_request(self.minio, self.job_request, self.db_dut)
        if job_bucket:
            job_bucket.create_owner_credentials("dut", groups=self.job_request.minio_groups,
                                                whitelisted_ips=[f'{self.db_dut.ip_address}/32'])
        return job_bucket

    @cached_property
    def job_httpd(self):
        return JobHTTPServer(self)

    @cached_property
    def job_config(self):
        # Bit nasty to render twice, but better than duplicating
        # template render in the various call-sites within
        # executor. Rendering it up front reduces the chances for
        # mistakes. (Meta-point: using an HTTP query to specify the
        # "target" could avoid this duplication of work, and might
        # actually make more sense)
        job = Job.render_with_resources(self.job_request.raw_job, self.db_dut, self.job_bucket,
                                        job={"http": {"url": self.job_httpd.url}})
        logger.debug("rendered job:\n%s", job)

        return job

    @cached_property
    def pdu_port(self):
        start = time.monotonic()

        config_pdu = self.mars_db.pdus.get(self.db_dut.pdu)
        if config_pdu is None:
            return None

        if pdu := PDU.create(config_pdu.driver, config_pdu.name, config_pdu.config, config_pdu.reserved_port_ids):
            for port in pdu.ports:
                if str(port.port_id) == str(self.db_dut.pdu_port_id):
                    port.min_off_time = self.db_dut.pdu_off_delay

                    exec_time = (time.monotonic() - start) * 1000
                    self.log(f"Initialized the PDU port in {exec_time:.1f} ms\n")

                    return port

            raise ValueError('Could not find a matching port for %s on %s' % (self.db_dut.pdu_port_id, pdu))

        raise ValueError("Could not create the PDU")

    def cancel_job(self):
        self.log("WARNING: The job got cancelled at the infra admin's request\n")
        self.stop_event.set()

    def log(self, msg, log_level=LogLevel.INFO):
        if not msg.endswith("\n"):
            msg += "\n"

        if self.job_console is not None:
            self.job_console.log(msg, log_level=log_level)

    def gen_job_boot_config(self, deployment):
        cache_mapping = self.remote_url_to_local_cache_mapping

        # Use the cached artifacts when possible, otherwise directly use the URL
        return BootConfig(kernel=cache_mapping.get(deployment.kernel_url, deployment.kernel_url),
                          initrd=cache_mapping.get(deployment.initramfs_url, deployment.initramfs_url),
                          dtb=cache_mapping.get(deployment.dtb_url, deployment.dtb_url),
                          cmdline=deployment.kernel_cmdline)

    def download_and_cache(self, url, output):
        attempts = 6
        tries = 0
        for i in range(attempts):
            try:
                tries += 1
                start_time = time.time()

                has_cache = config.EXECUTOR_ARTIFACT_CACHE_ROOT is not None
                if has_cache:
                    try:
                        os.makedirs(config.EXECUTOR_ARTIFACT_CACHE_ROOT, exist_ok=True)
                    except PermissionError:  # pragma: no cover
                        self.log(f"Unable to create \"{config.EXECUTOR_ARTIFACT_CACHE_ROOT}\", "
                                 "permission denied.  Disabling caching of artifacts...")
                        has_cache = False

                if has_cache:
                    session = CacheControl(requests.Session(), cache=FileCache(config.EXECUTOR_ARTIFACT_CACHE_ROOT))
                else:
                    session = requests.Session()

                resp = session.get(url)
                resp.raise_for_status()

                # Write the data to the output file
                output.seek(0, os.SEEK_SET)
                output.write(resp.content)
                output.flush()

                # Get the file size before resetting it to the beginning
                size_mb = output.tell() / 1024**2
                output.seek(0, os.SEEK_SET)

                # print some additional statistics when the fetch was with caching
                # disabled or was a cache miss
                if not has_cache or not resp.from_cache:
                    # note: don't round total_time here else it might cause a #DE when
                    # calculating avg speed
                    total_time = time.time() - start_time
                    avg_speed = size_mb / total_time
                    self.log(f"Downloaded {'and cached ' if has_cache else ''}"
                             f"the {url} in {round(total_time, 1)}s "
                             f"({round(size_mb, 2)} MB, at {round(avg_speed, 2)} MB/s)")
                else:
                    self.log(f"Fetched {url} from cache ({round(size_mb, 2)} MB)")

                return
            except Exception:
                self.log(traceback.format_exc())
                if i < attempts - 1:
                    self.log("Retrying to cache the artifact after a 10s delay\n")
                    time.sleep(10)
                else:
                    self.log("Failed too many times, aborting!\n")
                    self.job_console.set_state(JobConsoleState.OVER)

    def __fastboot_dut(self):
        def find_bootsdb_fastbootdevice():
            for path in [config.BOOTS_DB_USER_FILE, config.BOOTS_DB_FILE]:
                self.log(f"Trying to open the BootsDB {path}")
                if path and os.path.isfile(path):
                    try:
                        db = BootsDB.from_path(path, machine=self.db_dut)
                    except Exception:
                        self.log(f"WARNING: Could not parse the BOOTS DB file '{path}':\n{traceback.format_exc()}")
                        continue

                    for name, device in db.fastboot.items():
                        if device.match.matches(fbdev):
                            device.name = name
                            return device

        if self.cur_deployment:
            fbdev = FastbootDevice.from_serial(self.db_dut.id)
            if not fbdev:
                raise ValueError(f"Could not find the Fastboot device with serial {self.db_dut.id}")

            # If we found a device a matching fastboot device in boots db, enhance
            if db_fbdev := find_bootsdb_fastbootdevice():
                self.log(f"The Fastboot device matches BootsDB's {db_fbdev.name}: Applying defaults!")
                deployment = db_fbdev.defaults
                deployment.update(self.cur_deployment)
            else:
                deployment = self.cur_deployment

            # Ensure the boot configuration is valid
            cfg = self.gen_job_boot_config(deployment)
            if not cfg.kernel or not cfg.initrd or not cfg.dtb:
                self.log("ERROR: The generated boot configuration is missing a kernel, initrd, and/or an dtb")
                self.job_console.set_state(JobConsoleState.OVER)
                return cfg

            with tempfile.NamedTemporaryFile() as output:
                with tempfile.NamedTemporaryFile() as kernel:
                    with tempfile.NamedTemporaryFile() as initrd:
                        with tempfile.NamedTemporaryFile() as dtb:
                            args = [sys.executable, "-m", "valve_gfx_ci.executor.server.android.mkbootimg"]

                            if cfg.kernel:
                                self.download_and_cache(cfg.kernel, kernel)
                                args += ["--kernel", kernel.name, "--cmdline", cfg.cmdline]

                            if cfg.initrd:
                                self.download_and_cache(cfg.initrd, initrd)
                                args += ["--ramdisk", initrd.name]

                            if cfg.dtb:
                                self.download_and_cache(cfg.dtb, dtb)
                                args += ["--dtb", dtb.name]

                                # Append the DTB to the kernel image, as is expected by up to Android 9
                                # Source: https://source.android.com/docs/core/architecture/bootloader/dtb-images
                                with open(kernel.name, "ab") as f:
                                    f.write(dtb.read())
                                    f.flush()

                            if fastboot_cfg := deployment.fastboot:
                                for field, value in fastboot_cfg.fields_set.items():
                                    args += [f"--{field}", str(value)]

                            args += ["--output", output.name]

                            self.log(f"Generating a boot image using the following parameters: {args}")

                            subprocess.check_call(args, stdout=sys.stderr, stderr=sys.stderr)

                # Upload the generated boot image and run it
                self.log("Uploading the boot image")
                output.seek(0, os.SEEK_SET)
                fbdev.upload(output)

                # Booting the image
                self.log("Booting the image")
                fbdev.boot()

            return cfg
        else:
            return BootConfig()

    def boot_config_query(self, platform=None, buildarch=None, bootloader=None):
        self.log(f"The DUT queried its boot configuration as {bootloader} / {buildarch} / {platform}\n")

        if bootloader == "fastboot":
            return self.__fastboot_dut()
        elif self.cur_deployment:
            cfg = self.gen_job_boot_config(self.cur_deployment)
            cfg.fixup_missing_fields_with_defaults(platform=platform, buildarch=buildarch, bootloader=bootloader)

            if cfg.dtb and bootloader and bootloader not in ["uboot"]:
                self.log(f"WARNING: The `dtb` parameter is unsupported by {bootloader} and was thus ignored")

            return cfg

        return BootConfig()

    def _cache_remote_artifacts(self):
        def cache_it(url, cache_artifact_name):
            # If the URL is missing, don't try to cache it, we'll pick a default kernel/initrd based on the boot request
            if not url:
                self.remote_url_to_local_cache_mapping[url] = None
                return

            # Do not cache the URL if it is already found in our cache
            if url in self.remote_url_to_local_cache_mapping:
                return

            if self.minio.is_local_url(url):
                logger.debug(f"Ignore caching {url} as it is already hosted by our minio cache")
                return
            self.remote_url_to_local_cache_mapping[url] = f"{config.MINIO_URL}/boot/{cache_artifact_name}"

            self.log(f'Caching {url} into minio...\n')
            attempts = 6
            tries = 0
            for i in range(attempts):
                try:
                    tries += 1
                    self.minio.save_boot_artifact(url, f"{cache_artifact_name}", log_callback=self.log)
                    break
                except Exception:
                    self.log(traceback.format_exc())
                    if i < attempts - 1:
                        self.log("Retrying to cache the artifact after a 10s delay\n")
                        time.sleep(10)
                    else:
                        self.log("Failed too many times, aborting!\n")
                        self.job_console.set_state(JobConsoleState.OVER)
            self.log(f"Artifact fetched after {tries} attempt(s)\n")

        for url, paths in self.job_config.deployment.artifacts.items():
            for path, artifact in paths.items():
                path_str = "-".join((self.db_dut.id, ) + path)
                cache_it(url, path_str)

        if self.job_bucket:
            logger.info("Initializing the job bucket with the client's data")
            self.job_bucket.setup()

    def update_mars_fields(self, **fields):
        server_url = f"http://localhost:{config.EXECUTOR_PORT}/api/v1/dut/{self.db_dut.id}"
        r = requests.patch(server_url, json=fields)

        if r.status_code != 200:
            logger.error(f"ERROR: Failed to update the MaRS fields associated to this DUT. Reason: {r.text}")

    def run(self):
        def session_init():
            self.state = DUTState.RUNNING

            # Connect to the client's endpoint, to relay the serial console
            self.job_console = JobConsole(self.db_dut.id,
                                          client_endpoint=self.job_request.callback_endpoint,
                                          client_version=self.job_request.version)
            self.job_console.start(console_patterns=self.job_config.console_patterns)

            # Cut the power to the machine as early as possible, as we want to be
            # able to guarantee the power was off for the expected `min_off_time`,
            # and we can use some of that off time to setup the infra (download
            # kernel/initramfs, then push them to minio).
            self.pdu_port.set(PDUState.OFF)

        def session_end():
            # Ensure we cut the power to the DUT
            self.pdu_port.set(PDUState.OFF)

            self.job_console.close()
            self.job_console = None
            self.cur_deployment = None
            if self.job_bucket:
                self.job_bucket.remove()
                self.job_bucket = None

        def log_exception():
            logger.debug("Exception caught:\n%s", traceback.format_exc())
            self.log(f"An exception got caught: {traceback.format_exc()}\n", LogLevel.ERROR)
            # If exceptions start firing, throttle the parent loop, since it's
            # very heavy spam if left to run at full speed.
            time.sleep(2)

        def execute_job():
            # Start the overall timeout
            timeouts = self.job_config.timeouts
            timeouts.overall.start()

            # Download the kernel/initramfs
            self.log("Setup the infrastructure\n")
            timeouts.infra_setup.start()
            self._cache_remote_artifacts()
            self.log(f"Completed setup of the infrastructure, after {timeouts.infra_setup.active_for} s\n")
            timeouts.infra_setup.stop()

            # Keep on resuming until success, timeouts' retry limits is hit, or the entire executor is going down
            self.cur_deployment = self.job_config.deployment_start
            while (not self.stop_event.is_set() and
                   not timeouts.overall.has_expired and
                   self.job_console.state < JobConsoleState.DUT_DONE):
                self.job_console.reset_per_boot_state()

                # Make sure the machine shuts down
                self.pdu_port.set(PDUState.OFF)

                self.log(f"Power up the machine, enforcing {self.pdu_port.min_off_time} seconds of down time\n")
                self.pdu_port.set(PDUState.ON)

                # Start the boot, and enable the timeouts!
                self.log("Boot the machine\n")
                timeouts.boot_cycle.start()
                timeouts.first_console_activity.start()
                timeouts.console_activity.stop()

                # Reset all the watchdogs, since they are not supposed to remain active between rounds
                for wd in timeouts.watchdogs.values():
                    wd.cancel()

                while (self.job_console.state < JobConsoleState.DUT_DONE and
                       not self.job_console.needs_reboot and
                       not self.stop_event.is_set() and
                       not timeouts.has_expired):
                    # Update the activity timeouts, based on when was the
                    # last time we sent it data
                    if self.job_console.last_activity_from_machine is not None:
                        timeouts.first_console_activity.stop()
                        timeouts.console_activity.reset(when=self.job_console.last_activity_from_machine)

                    # Wait a little bit before checking again
                    time.sleep(0.1)

                # Cut the power
                self.pdu_port.set(PDUState.OFF)

                # Increase the retry count of the timeouts that expired, and
                # abort the job if we exceeded their limits.
                abort = False
                for timeout in timeouts.expired_list:
                    retry = timeout.retry()
                    decision = "Try again!" if retry else "Abort!"
                    self.log(f"Hit the timeout {timeout} --> {decision}\n", LogLevel.ERROR)
                    abort = abort or not retry

                # Check if the DUT asked us to reboot
                if self.job_console.needs_reboot:
                    retry = timeouts.boot_cycle.retry()
                    retries_str = f"{timeouts.boot_cycle.retried}/{timeouts.boot_cycle.retries}"
                    dec = f"Boot cycle {retries_str}, go ahead!" if retry else "Exceeded boot loop count, aborting!"
                    self.log(f"The DUT asked us to reboot: {dec}\n", LogLevel.WARN)
                    abort = abort or not retry

                if abort:
                    # We have reached a timeout retry limit, time to stop!
                    self.job_console.set_state(JobConsoleState.DUT_DONE)
                else:
                    # Stop all the timeouts, except the overall
                    timeouts.first_console_activity.stop()
                    timeouts.console_activity.stop()
                    timeouts.boot_cycle.stop()

                    # We went through one boot cycle, use the "continue" deployment
                    self.cur_deployment = self.job_config.deployment_continue

            # We either reached the end of the job, or the client got disconnected
            if self.job_console.state == JobConsoleState.DUT_DONE:
                # Mark the machine as unfit for service
                if self.job_console.machine_is_unfit_for_service:
                    self.log("The machine has been marked as unfit for service\n")
                    self.update_mars_fields(ready_for_service=False)

                # Tearing down the job
                self.log("The job has finished executing, starting tearing down\n")
                timeouts.infra_teardown.start()

                # Delay to make sure messages are read before the end of the job
                time.sleep(CONSOLE_DRAINING_DELAY)

                # Start the tear down, which will create and send the credentials
                # for the job bucket to the client
                self.log("Creating credentials to the job bucket for the client\n")
                self.job_console.set_state(JobConsoleState.TEAR_DOWN, job_bucket=self.job_bucket)

                # Wait for the client to close the connection
                self.log("Waiting for the client to download the job bucket\n")
                while (self.job_console.state < JobConsoleState.OVER and
                       not self.stop_event.is_set() and
                       not timeouts.infra_teardown.has_expired):
                    # Wait a little bit before checking again
                    time.sleep(0.1)

                self.log(f"Completed the tear down procedure in {timeouts.infra_teardown.active_for} s\n")
                timeouts.infra_teardown.stop()
            else:
                self.log("The job is over, skipping sharing the job bucket with the client")

            # We are done!

        try:
            session_init()
            self.log(f"Starting the job: {self.job_config}\n\n", LogLevel.DEBUG)
            execute_job()
        except Exception:
            log_exception()
        finally:
            session_end()

    @property
    def job_status(self):
        if self.job_console is not None:
            return self.job_console.console_patterns.job_status
        else:
            return JobStatus.UNKNOWN


app = flask.Flask(__name__)


def get_executor(raise_if_missing=True):
    with app.app_context():
        if executor := flask.current_app.executor:
            return executor
        elif raise_if_missing:
            raise ValueError("The executor has not started yet")


@app.errorhandler(ValueError)
def handle_valueError_exception(error):
    traceback.print_exc()
    response = flask.jsonify({"error": str(error)})
    response.status_code = 400
    return response


@app.route('/api/v1/state', methods=['GET'])
def get_state():
    if executor := get_executor(raise_if_missing=False):
        return {
            "state": executor.state.name
        }
    else:
        return {
            "state": DUTState.QUEUED.name
        }


@app.route('/api/v1/boot/config', methods=['GET'])
def get_boot_config():
    args = flask.request.args

    executor = get_executor()
    if boot_cfg := executor.boot_config_query(platform=args.get("platform"),
                                              buildarch=args.get("buildarch"),
                                              bootloader=args.get("bootloader")):
        return asdict(boot_cfg)


@app.route('/api/v1/job/cancel', methods=['POST'])
def cancel_job():
    executor = get_executor()
    executor.cancel_job()
    return flask.make_response("The job was marked for cancellation\n", 200)


@dataclass
class JobConfig:
    executor_job_version: int

    mars_db: MarsDB
    machine_id: str

    job_request: JobRequest

    pdu_port_stats: PduPortStats

    @field_validator("executor_job_version")
    @classmethod
    def executor_job_version_is_known(cls, v):
        assert v == 1
        return v


def run(config_f, socket_path, lock_path):
    def parse_config(config_f):
        cfg = yaml.safe_load(config_f)
        return JobConfig(**cfg)

    try:
        # Create an exclusive lock
        os.makedirs(os.path.dirname(f"{socket_path}"), exist_ok=True)
        os.makedirs(os.path.dirname(f"{lock_path}"), exist_ok=True)
        socket_lock = open(lock_path, "w")
        lock_fd(socket_lock.fileno())
        socket_lock.write(f"{os.getpid()}\n")
        socket_lock.flush()

        # Parse the configuration
        cfg = parse_config(config_f)
        db_dut = cfg.mars_db.duts.get(cfg.machine_id)
        if db_dut is None:
            raise ValueError(f"The machine id '{cfg.machine_id}' can't be found in mars_db")

        # HACK: We should really find a way to get this set directly by pydantic!
        db_dut.id = cfg.machine_id

        # Create the executor
        executor = Executor(mars_db=cfg.mars_db, db_dut=db_dut, job_request=cfg.job_request)

        # Copy the pdu port stats
        for f in fields(cfg.pdu_port_stats):
            setattr(executor.pdu_port, f.name, getattr(cfg.pdu_port_stats, f.name))

        # Update the configuration
        with app.app_context():
            flask.current_app.executor = executor
            flask.current_app.db_dut = db_dut

        # Disable Flask's access logging
        logging.getLogger('werkzeug').setLevel(logging.ERROR)

        # Start Flask
        flask_thread = threading.Thread(target=app.run, daemon=True,
                                        kwargs={"host": f"unix://{socket_path}",
                                                "port": None, "debug": True, "use_reloader": False})
        flask_thread.start()

        # Start the job
        executor.run()

        # Exit using the same status code as the job
        os._exit(executor.job_status.value)
    except Exception:
        # We caught an exception when we really shouldn't have, let's print it,
        # flush our streams, then die with the status code INCOMPLETE
        traceback.print_exc(file=sys.stderr)
        sys.stdout.flush()
        sys.stderr.flush()
        os._exit(JobStatus.INCOMPLETE.value)



from datetime import datetime, timedelta
from freezegun import freeze_time
from unittest import mock
from unittest.mock import patch, MagicMock
import os

import pytest
from pydantic import ValidationError

import server
from server.job import Target, Timeout, Timeouts, ConsoleState, Job, Pattern, CollectionOfLists
from server.job import Watchdog, KernelDeployment, DeploymentState, FastbootDeployment
import server.config as config

# Target


def test_Target_from_job__no_id_nor_tags():
    with pytest.raises(ValueError) as exc:
        Target()

    msg = "The target is neither identified by tags or id. Use empty tags to mean 'any machines'."
    assert msg in str(exc.value)


def test_Target_from_job__id_only():
    target_job = {
        "id": "MyID",
    }

    target = Target(**target_job)
    assert target.id == target_job['id']
    assert target.tags == []
    assert str(target) == f"<Target: id={target.id}, tags={target.tags}>"


def test_Target_from_job__tags_only():
    target_job = {
        "tags": ['tag 1', 'tag 2']
    }

    target = Target(**target_job)
    assert target.id is None
    assert target.tags == target_job['tags']
    assert str(target) == f"<Target: id={target.id}, tags={target.tags}>"


def test_Target_from_job__both_id_and_tags():
    target_job = {
        "id": "MyID",
        "tags": ['tag 1', 'tag 2']
    }

    target = Target(**target_job)
    assert target.id == target_job['id']
    assert target.tags == target_job['tags']
    assert str(target) == f"<Target: id={target.id}, tags={target.tags}>"


# Timeout


def test_Timeout__expiration_test():
    start_time = datetime(2021, 1, 1, 12, 0, 0)
    with freeze_time(start_time.isoformat()):
        timeout = Timeout(minutes=1, retries=0)
        assert timeout.started_at is None
        assert not timeout.is_started
        assert timeout.active_for is None

        # Start the timeout and check the state
        timeout.start()
        assert timeout.started_at == start_time
        assert timeout.is_started
        assert timeout.active_for == timedelta()
        assert not timeout.has_expired

        # Go right to the limit of the timeout
        delta = timedelta(seconds=60)
        with freeze_time((start_time + delta).isoformat()):
            assert timeout.started_at == start_time
            assert timeout.active_for == delta
            assert timeout.is_started
            assert not timeout.has_expired

        # And check that an extra millisecond trip it
        delta = timedelta(seconds=60, milliseconds=1)
        with freeze_time((start_time + delta).isoformat()):
            assert timeout.started_at == start_time
            assert timeout.active_for == delta
            assert timeout.is_started
            assert timeout.has_expired

        # Stop the timeout and check the state
        timeout.stop()
        assert timeout.started_at is None
        assert not timeout.is_started
        assert timeout.active_for is None


def test_Timeout__retry_lifecycle():
    timeout = Timeout(seconds=42, retries=1)

    # Check the default state
    assert timeout.started_at is None
    assert timeout.active_for is None
    assert timeout.retried == 0

    # Start the timeout
    start_time = datetime(2021, 1, 1, 12, 0, 0)
    with freeze_time(start_time.isoformat()):
        timeout.start()
        assert timeout.started_at == start_time
        assert timeout.retried == 0
        assert timeout.active_for == timedelta()
        assert not timeout.has_expired

    # Check that the default reset sets started_at to now()
    start_time = datetime(2021, 1, 1, 12, 0, 1)
    with freeze_time(start_time.isoformat()):
        timeout.reset()
        assert timeout.started_at == start_time

        # Check that a resetting to a certain time does as it should
        new_start = start_time - timedelta(seconds=1)
        timeout.reset(new_start)
        assert timeout.started_at == new_start

    # Do the first retry
    assert timeout.retry()
    assert timeout.started_at is None
    assert timeout.retried == 1

    # Second retry should fail
    assert not timeout.retry()


def test_Timeout_from_job():
    fields = {"days": 5, "hours": 6, "minutes": 7, "seconds": 8, "milliseconds": 9}

    delay = timedelta(**fields)
    timeout = Timeout(retries=42, **fields)
    timeout.name = "Yeeepeee"

    assert timeout.timeout == delay
    assert timeout.retries == 42
    assert str(timeout) == f"<Timeout Yeeepeee: value={delay}, retries=0/42>"


# Timeouts


def test_Timeouts__overall_with_retries():
    for t_type in ["overall", "infra_teardown"]:
        with pytest.raises(ValueError) as exc:
            Timeouts(**{t_type: Timeout(retries=1)})
        assert "Neither the overall nor the teardown timeout can have retries" in str(exc.value)


def test_Timeouts__default():
    timeouts = Timeouts()

    for timeout in timeouts:
        if timeout.name not in ["overall", "infra_teardown"]:
            assert timeout.timeout == timedelta.max
            assert timeout.retries == 0

    assert timeouts.expired_list == []
    assert not timeouts.has_expired
    assert timeouts.watchdogs == {}


def test_Timeouts__expired():
    overall = Timeout(days=1, retries=0)
    boot_cycle = Timeout(seconds=0, retries=0)
    wd1 = Timeout(seconds=0, retries=0)

    overall.start()
    boot_cycle.start()

    timeouts = Timeouts(overall=overall, boot_cycle=boot_cycle, watchdogs={"wd1": wd1})
    assert timeouts.has_expired
    assert timeouts.expired_list == [boot_cycle]

    boot_cycle.stop()
    assert not timeouts.has_expired
    assert timeouts.expired_list == []

    wd1.start()
    assert timeouts.has_expired
    assert timeouts.expired_list == [wd1]


def test_Timeouts__from_job():
    job_timeouts = {
        "first_console_activity": {
            "seconds": 45
        },
        "console_activity": {
            "seconds": 13
        },
        "watchdogs": {
            "custom1": {
                "seconds": 42
            }
        }
    }

    timeouts = Timeouts(**job_timeouts)

    assert timeouts.first_console_activity.timeout == timedelta(seconds=45)
    assert timeouts.console_activity.timeout == timedelta(seconds=13)
    assert timeouts.watchdogs.get("custom1").timeout == timedelta(seconds=42)
    assert timeouts.watchdogs["custom1"] in timeouts


# Pattern


def test_Pattern_from_job__invalid_regex():
    with pytest.raises(ValueError) as excinfo:
        Pattern(**{"regex": "BOOM\\"})

    error_msg = "Console pattern 'BOOM\\' is not a valid regular expression: bad escape (end of pattern)"
    assert error_msg in str(excinfo.value)


# Watchdogs


def test_Watchdog__process_line():
    wd = Watchdog(**{
        "start": {"regex": "start"},
        "reset": {"regex": "reset"},
        "stop": {"regex": "stop"},
    })

    # Check that nothing explodes if we have no timeouts set
    assert wd.process_line(b"line") == {}

    # Set the timeout
    wd.set_timeout(MagicMock(is_started=False))
    wd.timeout.start.assert_not_called()
    wd.timeout.reset.assert_not_called()
    wd.timeout.stop.assert_not_called()

    # Check that sending the reset/stop patterns before starting does nothing
    assert wd.process_line(b"line reset line") == {}
    assert wd.process_line(b"line stop line") == {}
    wd.timeout.start.assert_not_called()
    wd.timeout.reset.assert_not_called()
    wd.timeout.stop.assert_not_called()

    # Check that the start pattern starts the timeout
    assert wd.process_line(b"line start line") == {"start"}
    wd.timeout.start.assert_called_once()
    wd.timeout.reset.assert_not_called()
    wd.timeout.stop.assert_not_called()

    # Emulate the behaviour of the timeout
    wd.timeout.is_started = True

    # Check that the start pattern does not restart the timeout
    assert wd.process_line(b"line start line") == {}
    wd.timeout.start.assert_called_once()
    wd.timeout.reset.assert_not_called()
    wd.timeout.stop.assert_not_called()

    # Check that the reset pattern works
    assert wd.process_line(b"line reset line") == {"reset"}
    wd.timeout.start.assert_called_once()
    wd.timeout.reset.assert_called_once()
    wd.timeout.stop.assert_not_called()

    # Check that the stop pattern works
    assert wd.process_line(b"line stop line") == {"stop"}
    wd.timeout.start.assert_called_once()
    wd.timeout.reset.assert_called_once()
    wd.timeout.stop.assert_called_once()


def test_Watchdog__stop():
    wd = Watchdog(**{
        "start": {"regex": "start"},
        "reset": {"regex": "reset"},
        "stop": {"regex": "stop"},
    })

    # Check that nothing explodes if we have no timeouts set
    wd.cancel()

    # Set the timeout
    wd.set_timeout(MagicMock(is_started=False))
    wd.timeout.stop.assert_not_called()

    # Check that sending the reset/stop patterns before starting does nothing
    wd.cancel()
    wd.timeout.stop.assert_called_once()


# ConsoleState


def test_ConsoleState__missing_session_end():
    with pytest.raises(ValidationError):
        ConsoleState(session_end=None, session_reboot=None, job_success=None, job_warn=None,
                     machine_unfit_for_service=None)


def test_ConsoleState__simple_lifecycle():
    state = ConsoleState(session_end=Pattern("session_end"), session_reboot=None, job_success=None, job_warn=None,
                         machine_unfit_for_service=None)

    assert state.job_status == "INCOMPLETE"
    assert not state.session_has_ended
    assert not state.needs_reboot

    state.process_line(b"oh oh oh")
    assert state.job_status == "INCOMPLETE"
    assert not state.session_has_ended
    assert not state.needs_reboot

    state.process_line(b"blabla session_end blaba\n")
    assert state.job_status == "COMPLETE"
    assert state.session_has_ended
    assert not state.needs_reboot


def test_ConsoleState__lifecycle_with_extended_support():
    state = ConsoleState(session_end=Pattern("session_end"), session_reboot=Pattern("session_reboot"),
                         job_success=Pattern("job_success"), job_warn=Pattern("job_warn"),
                         machine_unfit_for_service=Pattern("machine_unfit_for_service"),
                         watchdogs={"wd1": Watchdog(start=Pattern(r"wd1_start"),
                                                    reset=Pattern(r"wd1_reset"),
                                                    stop=Pattern(r"wd1_stop"))})

    assert state.job_status == "INCOMPLETE"
    assert not state.session_has_ended
    assert not state.needs_reboot
    assert not state.machine_is_unfit_for_service

    assert state.process_line(b"oh oh oh") == set()
    assert state.job_status == "INCOMPLETE"
    assert not state.session_has_ended
    assert not state.needs_reboot
    assert not state.machine_is_unfit_for_service

    assert state.process_line(b"blabla session_reboot blabla") == {"session_reboot"}
    assert state.job_status == "INCOMPLETE"
    assert not state.session_has_ended
    assert state.needs_reboot
    assert not state.machine_is_unfit_for_service

    state.reset_per_boot_state()
    assert not state.session_has_ended
    assert not state.needs_reboot

    assert state.process_line(b"blabla session_end blaba\n") == {"session_end"}
    assert state.job_status == "FAIL"
    assert state.session_has_ended
    assert not state.needs_reboot
    assert not state.machine_is_unfit_for_service

    assert state.process_line(b"blabla job_success blaba\n") == {"job_success"}
    assert state.job_status == "PASS"
    assert state.session_has_ended
    assert not state.needs_reboot
    assert not state.machine_is_unfit_for_service

    assert state.process_line(b"blabla job_warn blaba\n") == {"job_warn"}
    assert state.job_status == "WARN"
    assert state.session_has_ended
    assert not state.needs_reboot
    assert not state.machine_is_unfit_for_service

    assert state.process_line(b"blabla machine_unfit_for_service blaba\n") == {"machine_unfit_for_service"}
    assert state.job_status == "WARN"
    assert state.session_has_ended
    assert not state.needs_reboot
    assert state.machine_is_unfit_for_service

    state.watchdogs.get("wd1").set_timeout(Timeout.create(name="wd1", seconds=1, retries=1))
    assert state.process_line(b"blabla wd1_start blaba\n") == {"wd1.start"}
    assert state.job_status == "WARN"
    assert state.session_has_ended
    assert not state.needs_reboot
    assert state.machine_is_unfit_for_service


def test_ConsoleState__default():
    console_state = ConsoleState()

    print(console_state.session_end.regex.pattern)
    assert console_state.session_end.regex.pattern == b"^\\[[\\d \\.]{12}\\] reboot: Power Down$"
    assert console_state.session_reboot is None
    assert console_state.job_success is None
    assert console_state.job_warn is None
    assert console_state.machine_unfit_for_service is None

    config.CONSOLE_PATTERN_DEFAULT_MACHINE_UNFIT_FOR_SERVICE_REGEX = "TEST"
    console_state = ConsoleState()
    assert console_state.machine_unfit_for_service.regex.pattern == b"TEST"


def test_ConsoleState_from_job__full():
    console_state = ConsoleState(**{
        "session_end": {
            "regex": "session_end"
        }, "session_reboot": {
            "regex": "session_reboot"
        }, "job_success": {
            "regex": "job_success"
        }, "job_warn": {
            "regex": "job_warn"
        }, "machine_unfit_for_service": {
            "regex": "unfit_for_service"
        }, "watchdogs": {
            "mywatchdog": {
                "start": {"regex": "start"},
                "reset": {"regex": "reset"},
                "stop": {"regex": "stop"},
            }
        }
    })

    assert console_state.session_end.regex.pattern == b"session_end"
    assert console_state.session_reboot.regex.pattern == b"session_reboot"
    assert console_state.job_success.regex.pattern == b"job_success"
    assert console_state.job_warn.regex.pattern == b"job_warn"
    assert console_state.machine_unfit_for_service.regex.pattern == b"unfit_for_service"


# CollectionOfLists


def test_CollectionOfLists__invalid_collection_name():
    # Can't use a keyword as category
    with pytest.raises(ValidationError):
        CollectionOfLists[str]({":uncategorised": ["Hello"]})


def test_CollectionOfLists__from_job():
    # CollectionOfLists
    val = CollectionOfLists[str]()
    assert CollectionOfLists[str].from_job(val) == val

    # Single value
    assert CollectionOfLists[str].from_job("hello").as_list == ["hello"]

    # List of values
    assert CollectionOfLists[int].from_job([1, 2]).as_list == [1, 2]

    # Categories
    cl = CollectionOfLists[str].from_job({"01-cat1": "Neo!",
                                          "00-cat2": ["Wake", "up,"],
                                          ":uncategorised": ["Don't", "be", "lazy!"]})
    assert cl.uncategorised == ["Don't", "be", "lazy!"]
    assert cl.as_list == ["Wake", "up,", "Neo!", "Don't", "be", "lazy!"]
    assert list(cl) == cl.as_list

    # Invalid collection name
    with pytest.raises(ValidationError):
        CollectionOfLists[str].from_job({":cat": ["Hello"]})


def test_CollectionOfLists__str():
    assert str(CollectionOfLists[str](uncategorised=["Wake", "up,", "Neo!"])) == "Wake up, Neo!"


def test_CollectionOfLists__equality():
    categories = {
        "cat1": ["Hello"],
        "cat2": ["World"]
    }
    uncategorised = ["Wake", "up,", "Neo!"]

    # Make sure that 2 objects constructed with the same values are equal
    assert CollectionOfLists[str](categories, uncategorised) == CollectionOfLists[str](categories, uncategorised)

    # Make sure all combinaisons of missing inputs don't result in a match
    assert CollectionOfLists[str](categories) != CollectionOfLists[str](categories, uncategorised)
    assert CollectionOfLists[str](uncategorised=uncategorised) != CollectionOfLists[str](categories, uncategorised)
    assert CollectionOfLists[str](categories, uncategorised) != CollectionOfLists[str](categories)
    assert CollectionOfLists[str](categories, uncategorised) != CollectionOfLists[str](uncategorised=uncategorised)

    # Comparison to a string
    assert CollectionOfLists[str](categories, uncategorised) == "Hello World Wake up, Neo!"


def test_CollectionOfLists__update():
    categories = {
        "greeting": ["Hello"],
        "who": ["World"],
    }
    uncategorised = ["and", "the", "universe!"]

    c = CollectionOfLists[str](categories=categories, uncategorised=uncategorised)
    assert c == "Hello World and the universe!"

    # Replacing a category also keeps the uncategorised
    c.update(CollectionOfLists[str](categories={"greeting": ["Bye Bye"]}))
    assert c == "Bye Bye World and the universe!"

    # Get rid of a complete category, add another one, and replace the uncategorised list
    c.update(CollectionOfLists[str](categories={"greeting": ["Welcome"], "who": []},
                                    uncategorised=["my", "son", "to", "the", "machine"]))
    assert c == "Welcome my son to the machine"


# Deployment

def test_KernelDeployment_multiline_string():
    assert KernelDeployment(cmdline="toto").cmdline == "toto"
    assert KernelDeployment(cmdline=["tag1", "tag2"]).cmdline == "tag1 tag2"
    assert KernelDeployment().cmdline is None


def test_FastbootDeployment():
    assert str(FastbootDeployment()) == "<Fastboot: >"

    fb = str(FastbootDeployment(header_version=4, dtb_offset=0x123456))
    assert fb == "<Fastboot: header_version=4, dtb_offset=0x123456>"


def test_DeploymentState():
    deployment = DeploymentState()

    assert deployment.kernel_url is None
    assert deployment.initramfs_url is None
    assert deployment.kernel_cmdline is None

    deployment.update({})

    assert deployment.kernel_url is None
    assert deployment.initramfs_url is None
    assert deployment.kernel_cmdline is None

    deployment.update({"kernel": {"url": "https://host.tld/kernel_url", "cmdline": "cmdline"},
                       "initramfs": {"url": "https://host.tld/initramfs_url"},
                       "dtb": {"url": "https://host.tld/dtb_url"}})

    assert deployment.kernel_url == "https://host.tld/kernel_url"
    assert deployment.initramfs_url == "https://host.tld/initramfs_url"
    assert deployment.kernel_cmdline == "cmdline"

    assert str(deployment) == """<Deployment:
    kernel_url: https://host.tld/kernel_url
    initramfs_url: https://host.tld/initramfs_url
    kernel_cmdline: cmdline
    dtb_url: https://host.tld/dtb_url>
"""

    # Update only one kernel field
    deployment.update({"kernel": {"cmdline": "cmdline2"}})
    assert deployment.kernel_url == "https://host.tld/kernel_url"
    assert deployment.initramfs_url == "https://host.tld/initramfs_url"
    assert deployment.kernel_cmdline == "cmdline2"


# Job


def test_Job__simple():
    simple_job = """
version: 1
target:
  id: "b4:2e:99:f0:76:c5"
console_patterns:
  session_end:
    regex: "session_end"
deployment:
  start:
    kernel:
      url: "https://host.tld/kernel_url"
      cmdline:
        - my
        - start cmdline
    initramfs:
      url: "https://host.tld/initramfs_url"
"""
    job = Job.render_with_resources(simple_job)

    assert job.version == 1
    assert job.deadline == datetime.max

    assert job.target.id == "b4:2e:99:f0:76:c5"
    assert job.target.tags == []

    assert job.deployment_start.kernel_url == "https://host.tld/kernel_url"
    assert job.deployment_start.initramfs_url == "https://host.tld/initramfs_url"
    assert job.deployment_start.kernel_cmdline == "my start cmdline"
    assert job.deployment_start.fastboot is None
    assert job.deployment_start.artifacts == {
        "https://host.tld/kernel_url": {("kernel", ): job.deployment_start.kernel},
        "https://host.tld/initramfs_url": {("initramfs", ): job.deployment_start.initramfs},
    }

    assert job.deployment_continue.kernel_url == job.deployment_start.kernel_url
    assert job.deployment_continue.initramfs_url == job.deployment_start.initramfs_url
    assert job.deployment_continue.kernel_cmdline == job.deployment_start.kernel_cmdline
    assert job.deployment_continue.fastboot == job.deployment_start.fastboot
    assert job.deployment_continue.artifacts == job.deployment_start.artifacts

    assert job.deployment.artifacts == {
        "https://host.tld/kernel_url": {
            ("start", "kernel", ): job.deployment_start.kernel,
            ("continue", "kernel", ): job.deployment_continue.kernel
        },
        "https://host.tld/initramfs_url": {
            ("start", "initramfs", ): job.deployment_start.initramfs,
            ("continue", "initramfs", ): job.deployment_continue.initramfs
        },
    }

    # Make sure the job's __str__ method does not crash
    str(job)


def test_Job__override_continue():
    override_job = """
version: 1
deadline: "2021-03-31 00:00:00"
target:
  id: "b4:2e:99:f0:76:c6"
  tags: ["amdgpu:gfxversion::gfx10"]
console_patterns:
  session_end:
    regex: "session_end"
deployment:
  start:
    kernel:
      url: "https://host.tld/kernel_url"
      cmdline:
        defaults:
          - my
          - default cmdline
    initramfs:
      url: "https://host.tld/initramfs_url"
  continue:
    kernel:
      url: "https://host.tld/kernel_url_2"
      cmdline: "my continue cmdline"
    initramfs:
      url: "https://host.tld/initramfs_url_2"
    dtb:
      url: "https://host.tld/dtb_url_2"
    fastboot:
      header_version: 42
      base: 0x123456789
      kernel_offset: 0xcafe
      ramdisk_offset: 0xc0de
      dtb_offset: 0xbeef
      tags_offset: 0xdead
      board: "myboard"
      pagesize: 16384
"""
    job = Job.render_with_resources(override_job)

    assert job.version == 1
    assert job.deadline == datetime.fromisoformat("2021-03-31 00:00:00")

    assert job.target.id == "b4:2e:99:f0:76:c6"
    assert job.target.tags == ["amdgpu:gfxversion::gfx10"]

    assert job.deployment_start.kernel_url == "https://host.tld/kernel_url"
    assert job.deployment_start.initramfs_url == "https://host.tld/initramfs_url"
    assert job.deployment_start.kernel_cmdline == "my default cmdline"
    assert job.deployment_start.dtb_url is None
    assert job.deployment_start.artifacts == {
        "https://host.tld/kernel_url": {("kernel", ): job.deployment_start.kernel},
        "https://host.tld/initramfs_url": {("initramfs", ): job.deployment_start.initramfs},
    }

    assert job.deployment_continue.kernel_url == "https://host.tld/kernel_url_2"
    assert job.deployment_continue.initramfs_url == "https://host.tld/initramfs_url_2"
    assert job.deployment_continue.kernel_cmdline == "my default cmdline my continue cmdline"
    assert job.deployment_continue.dtb_url == "https://host.tld/dtb_url_2"
    assert job.deployment_continue.artifacts == {
        "https://host.tld/kernel_url_2": {("kernel", ): job.deployment_continue.kernel},
        "https://host.tld/initramfs_url_2": {("initramfs", ): job.deployment_continue.initramfs},
        "https://host.tld/dtb_url_2": {("dtb", ): job.deployment_continue.dtb},
    }

    assert str(job.deployment_continue.fastboot) == ("<Fastboot: header_version=42, base=0x123456789, "
                                                     "kernel_offset=0xcafe, ramdisk_offset=0xc0de, dtb_offset=0xbeef, "
                                                     "tags_offset=0xdead, board=myboard, pagesize=16384>")

    assert job.deployment.artifacts == {
        "https://host.tld/kernel_url": {("start", "kernel", ): job.deployment_start.kernel},
        "https://host.tld/initramfs_url": {("start", "initramfs", ): job.deployment_start.initramfs},
        "https://host.tld/kernel_url_2": {("continue", "kernel", ): job.deployment_continue.kernel},
        "https://host.tld/initramfs_url_2": {("continue", "initramfs", ): job.deployment_continue.initramfs},
        "https://host.tld/dtb_url_2": {("continue", "dtb", ): job.deployment_continue.dtb},
    }


class MockMachine:
    @property
    def ready_for_service(self):
        return True

    @property
    def id(self):
        return "b4:2e:99:f0:76:c5"

    @property
    def tags(self):
        return ["some", "tags"]

    @property
    def local_tty_device(self):
        return "ttyS0"

    @property
    def ip_address(self):
        return "10.42.0.123"

    @property
    def safe_attributes(self):
        return {
            "base_name": "base_name",
            "full_name": "full_name",
            "tags": self.tags,
            "ip_address": self.ip_address,
            "local_tty_device": self.local_tty_device,
            "ready_for_service": self.ready_for_service,
        }


class MockBucket:
    @property
    def name(self):
        return "bucket_name"

    @property
    def minio(self):
        return MagicMock(url="minio_url")

    @property
    def credentials(self):
        return MagicMock(dut=("access", "secret"))


@patch('server.config.job_environment_vars')
def test_Job__sample(job_env):
    job_env.return_value = {'MINIO_URL': 'http://localhost:9000/testing-url',
                            'NTP_PEER': '10.42.0.1',
                            'PULL_THRU_REGISTRY': '10.42.0.1:8001'}

    m = MockMachine()
    job = Job.from_path("src/valve_gfx_ci/executor/server/tests/sample_job.yml", m)

    assert job.version == 1
    assert job.deadline == datetime.fromisoformat("2021-03-31 00:00:00")

    assert job.target.id == m.id
    assert job.target.tags == m.tags

    assert job.deployment_start.kernel_url == "http://localhost:9000/testing-url/test-kernel"
    assert job.deployment_start.initramfs_url == "http://localhost:9000/testing-url/test-initramfs"

    assert job.deployment_start.kernel_cmdline == 'b2c.container="docker://10.42.0.1:8001/infra/machine-registration:latest check" b2c.ntp_peer="10.42.0.1" b2c.pipefail b2c.cache_device=auto b2c.container="-v /container/tmp:/storage docker://10.42.0.1:8002/tests/mesa:12345" console=ttyS0,115200 earlyprintk=vga,keep SALAD.machine_id=b4:2e:99:f0:76:c5 extra=""'  # noqa: E501

    assert job.deployment_continue.kernel_url == "http://localhost:9000/testing-url/test-kernel"
    assert job.deployment_continue.initramfs_url == "http://localhost:9000/testing-url/test-initramfs"
    assert job.deployment_continue.kernel_cmdline == 'b2c.container="docker://10.42.0.1:8001/infra/machine-registration:latest check" b2c.ntp_peer=10.42.0.1 b2c.pipefail b2c.cache_device=auto b2c.container="-v /container/tmp:/storage docker://10.42.0.1:8002/tests/mesa:12345 resume"'  # noqa: E501


def test_Job__invalid_format():
    job = """
version: 1
target:
  id: "b4:2e:99:f0:76:c6"
console_patterns:
  session_end:
    regex: "session_end"
  reboot:
    regex: "toto"
deployment:
  start:
    kernel:
      url: "https://host.tld/kernel_url"
      cmdline:
        - my
        - start cmdline
    initramfs:
      url: "https://host.tld/initramfs_url"
"""

    with pytest.raises(ValueError) as exc:
        Job.render_with_resources(job)

    assert "console_patterns.reboot\n  Unexpected keyword argument" in str(exc.value)


@patch('server.config.job_environment_vars')
def test_Job__from_machine(job_env):
    job_env.return_value = {'NTP_PEER': '10.42.0.1'}

    simple_job = """
version: 1
target:
  id: {{ machine_id }}
console_patterns:
  session_end:
    regex: "session_end"
deployment:
  start:
    kernel:
      url: "https://host.tld/kernel_url"
      cmdline:
        - my {{ minio_url }}
        - start cmdline {{ ntp_peer }}
        - hostname {{ machine.full_name }}
        - extra {{ extra }}
    initramfs:
      url: "https://host.tld/initramfs_url"
"""
    job = Job.render_with_resources(simple_job, MockMachine(), MockBucket(), extra="parameter")

    assert job.version == 1
    assert job.deadline == datetime.max

    assert job.target.id == "b4:2e:99:f0:76:c5"
    assert job.target.tags == []

    assert job.deployment_start.kernel_url == "https://host.tld/kernel_url"
    assert job.deployment_start.initramfs_url == "https://host.tld/initramfs_url"
    assert (job.deployment_start.kernel_cmdline == "my minio_url start cmdline 10.42.0.1 hostname full_name "
                                                   "extra parameter")

    assert job.deployment_continue.kernel_url == job.deployment_start.kernel_url
    assert job.deployment_continue.initramfs_url == job.deployment_start.initramfs_url
    assert job.deployment_continue.kernel_cmdline == job.deployment_start.kernel_cmdline


def test_Job__watchdogs():
    override_job = """
version: 1
target:
  id: "b4:2e:99:f0:76:c6"
timeouts:
  watchdogs:
    wd1:
      minutes: 1
console_patterns:
  session_end:
    regex: "session_end"
  watchdogs:
    wd1:
      start:
        regex: "start"
      reset:
        regex: "reset"
      stop:
        regex: "stop"
deployment:
  start:
    kernel:
      url: "https://host.tld/kernel_url"
      cmdline: "cmdline"
    initramfs:
      url: "https://host.tld/initramfs_url"
"""
    job = Job.render_with_resources(override_job)
    assert job.console_patterns.watchdogs["wd1"].timeout == job.timeouts.watchdogs["wd1"]

    # Test that getting the string does not explode
    str(job)


# Job vars

@mock.patch.dict(os.environ, {"EXECUTOR_JOB__FDO_PROXY_REGISTRY": "10.10.10.1:1234"})
def test_server_config_job_environment_vars():
    ret = server.config.job_environment_vars()

    assert "MINIO_URL" in ret

    assert "FDO_PROXY_REGISTRY" in ret
    assert ret["FDO_PROXY_REGISTRY"] == "10.10.10.1:1234"



from dataclasses import dataclass
from urllib.request import urlretrieve
from urllib.parse import urlparse
from threading import Thread
from tempfile import TemporaryFile
from typing import Optional
import hashlib
import socket
import os
import logging
import struct
import binascii
import time
import re

from . import config
from .android.fastbootd import Fastbootd
from .tftpd import TFTPD, TFTPDClient
from .dhcpd import DHCPD, CPUArch, Firmware, BootProtocol, MacAddress
from .logger import logger
from .socketactivation import get_sockets_by_name

DEFAULT_CONFIG_PATHS = {
    'TFTP_DIR': '/boots/tftp',
}

# The IPXE binaries can get generated using "make ipxe-dut-clients"
IPXE_BASE_URL = 'https://downloads.gfx-ci.steamos.cloud/ipxe-dut-client/'
IPXE_BINARIES = {
    "IPXE_X86_64_EFI_FILENAME": '2024-01-30_20-41-29-mupuf-x86_64-snponly.efi',
    "IPXE_I386_EFI_FILENAME": '2024-01-30_20-41-29-mupuf-i386-snponly.efi',
    "IPXE_I386_MBR_FILENAME": '2024-01-30_20-41-29-mupuf-i386-undionly.kpxe',
    "IPXE_ARM32_EFI_FILENAME": '2024-01-30_20-41-29-mupuf-arm32-snponly.efi',
    "IPXE_ARM64_EFI_FILENAME": '2024-01-30_20-41-29-mupuf-arm64-snponly.efi',
}
BASE_DIR = os.path.dirname(__file__)


def provision_ipxe_dut_clients(tftp_dir):  # pragma: nocover
    os.makedirs(tftp_dir, exist_ok=True)

    logger.debug("Downloading the latest iPXE DUT clients...")
    for filename in IPXE_BINARIES.values():
        urlretrieve(f"{IPXE_BASE_URL}/{filename}", os.path.join(tftp_dir, filename))


class BootsDHCPD(DHCPD, Thread):
    def __init__(self, boots, name, interface):  # pragma: nocover
        self.boots = boots

        Thread.__init__(self, name=name)
        self.daemon = True

        DHCPD.__init__(self, interface=interface)

    @property
    def static_clients(self):
        clients = []
        for dut in self.boots.mars.known_machines:
            clients.append({'mac_addr': MacAddress(dut.mac_address),
                            'ipaddr': dut.ip_address,
                            'hostname': dut.full_name})
        return clients

    def boot_target(self, client_request):
        # Ensure MaRS is aware of this client request
        mac_addr = client_request.mac_addr
        ip_address, _ = self.get_or_assign_ip_for_client(mac_addr)
        arch = client_request.architecture.name
        fw = client_request.firmware.name
        protocol = client_request.protocol.name
        dut_data = {
            "id": mac_addr.as_str,
            "mac_address": mac_addr.as_str,
            "base_name": f"{arch}-{fw}-{protocol}".lower(),
            "ip_address": ip_address,
            "tags": []
        }
        self.boots.mars.machine_discovered(dut_data, update_if_already_exists=False)

        if client_request.architecture not in [CPUArch.X86, CPUArch.X86_64, CPUArch.ARM32, CPUArch.ARM64]:
            self.logger.error(f"Unsupported architecture '{client_request.architecture}'")
            return None

        if client_request.protocol != BootProtocol.TFTP:
            self.logger.error(f"Unsupported protocol '{client_request.protocol}'")
            return None

        if client_request.user_class == "iPXE":
            p = client_request.firmware.to_ipxe_platform if client_request.firmware else ""
            b = client_request.architecture.to_ipxe_buildarch if client_request.architecture else ""
            return f"{config.EXECUTOR_URL}/boot/{client_request.mac_addr}/boot.ipxe?platform={p}&buildarch={b}"
        elif client_request.firmware == Firmware.BIOS:
            if client_request.architecture in [CPUArch.X86, CPUArch.X86_64]:
                # NOTE: BIOS firmware means an x86-compatible CPU, and the i386 binary will work for all of them
                return IPXE_BINARIES['IPXE_I386_MBR_FILENAME']
            else:
                self.logger.error(f"Unsupported BIOS architecture '{client_request.architecture}'")
        elif client_request.firmware == Firmware.UEFI:
            if client_request.architecture == CPUArch.X86:
                return IPXE_BINARIES['IPXE_I386_EFI_FILENAME']
            elif client_request.architecture == CPUArch.X86_64:
                return IPXE_BINARIES['IPXE_X86_64_EFI_FILENAME']
            elif client_request.architecture == CPUArch.ARM32:
                return IPXE_BINARIES['IPXE_ARM32_EFI_FILENAME']
            elif client_request.architecture == CPUArch.ARM64:
                return IPXE_BINARIES['IPXE_ARM64_EFI_FILENAME']
            else:  # pragma: nocover
                self.logger.error(f"Unsupported UEFI architecture '{client_request.architecture}'")
        elif client_request.firmware == Firmware.UBOOT:
            arch = client_request.architecture.name
            return f"/dyn/bootloader/u-boot/{arch}/{client_request.mac_addr}/boot.scr"
        else:
            self.logger.error(f"Unsupported firmware type '{client_request.firmware}'")
            return None

    def run(self):  # pragma: nocover
        # Use the socket provided by our caller (systemd?), or create or own if none are found
        sock = None
        if sockets := get_sockets_by_name(config.BOOTS_DHCP_IPv4_SOCKET_NAME, socket.AF_INET, socket.SOCK_DGRAM):
            sock = sockets[0]

        self.listen(sock)


class TftpdClient(TFTPDClient):  # pragma: nocover
    def load_file(self, filename):
        if m := re.fullmatch("/?dyn/bootloader/u-boot/([^/]*)/([^/]*)/boot.scr", filename):
            buildarch = m.group(1)
            machine_id = m.group(2)
            data = self.parent.boots.uboot_boot_script(buildarch, machine_id)

            try:
                self.fh = TemporaryFile()
                self.fh.write(data)
                self.fh.seek(0)

                self.filesize = len(data)
                self.filename = 'gen://' + filename
            except Exception as e:
                self.parent.logger.warning(f"While serving {filename} caught the following exception: {e}")
                return False

            return True

        if m := re.fullmatch("/?minio(/.*)", filename):
            minio_url = config.MINIO_URL + m.group(1)
            try:
                tempfile, headers = urlretrieve(minio_url)
            except Exception as e:
                self.parent.logger.warning(f"While fetching {minio_url} caught the following exception: {e}")
                return False

            try:
                self.fh = open(tempfile, 'rb')

                self.fh.seek(0, os.SEEK_END)
                self.filesize = self.fh.tell()
                self.fh.seek(0, os.SEEK_SET)

                self.filename = "minio://" + filename
            except Exception as e:
                self.parent.logger.warning(f"While serving {minio_url} caught the following exception: {e}")
                return False
            finally:
                os.remove(tempfile)

            return True

        return TFTPDClient.load_file(self, filename)


class BootsTFTPD(TFTPD, Thread):  # pragma: nocover
    def __init__(self, boots, name, directory, interface):
        self.boots = boots

        Thread.__init__(self, name=name)
        self.daemon = True

        tftp_logger = logging.getLogger(f"{logger.name}.TFTP")
        tftp_logger.setLevel(logging.INFO)
        TFTPD.__init__(self, interface, client=TftpdClient, logger=tftp_logger, netboot_directory=directory)

    def run(self):
        # Use the socket provided by our caller (systemd?), or create or own if none are found
        sock = None
        if sockets := get_sockets_by_name(config.BOOTS_TFTP_IPv4_SOCKET_NAME, socket.AF_INET, socket.SOCK_DGRAM):
            sock = sockets[0]

        self.listen(sock)


class BootsFastbootd(Fastbootd, Thread):  # pragma: nocover
    def __init__(self, boots, name):
        self.boots = boots

        Thread.__init__(self, name=name)
        self.daemon = True

        fastbootd_logger = logging.getLogger(f"{logger.name}.Fastbootd")
        fastbootd_logger.setLevel(logging.INFO)
        Fastbootd.__init__(self, logger=fastbootd_logger)

    def gen_mac_address(self, serial):
        mac_addr = bytearray(hashlib.sha1(serial.encode()).digest()[-7:-1])

        # Ensure we generated a locally-administrated MAC address
        if (mac_addr[0] & 0x0F) not in [0x2, 0x6, 0xA, 0xE]:
            mac_addr[0] = (mac_addr[0] & 0x0F) | 0x2

        return MacAddress(bytes(mac_addr))

    def new_device_found(self, device):  # pragma: nocover
        # Ensure the bootloader is unlocked
        if device.variables.get("unlocked") != "yes":
            try:
                device.run_cmd(b"flashing:unlock")
            except Exception:
                # Ignore errors, we'll just tell users that the device is unusable
                pass

        # Ensure MaRS is aware of this client request
        serialno = device.variables["serialno"]
        mac_addr = self.gen_mac_address(serialno)
        product_name = device.variables.get("product")
        base_name = product_name or serialno
        ip_address, _ = self.boots.dhcpd.get_or_assign_ip_for_client(mac_addr)
        dut_data = {
            "id": serialno,
            "mac_address": mac_addr.as_str,
            "base_name": f"fastboot-{base_name}".lower(),
            "ip_address": ip_address,
            "tags": []
        }
        self.boots.mars.machine_discovered(dut_data, update_if_already_exists=False)

        # We got everything we wanted from the device, free our lock
        device.release()

        # Notify the DUT that we are ready to boot!
        if dut := self.boots.mars.get_machine_by_id(serialno):
            dut.boot_config_query(platform=None, buildarch=None, bootloader="fastboot")

    def run(self):
        while True:
            self.update_device_list()
            time.sleep(1)


@dataclass
class BootConfig:
    kernel: Optional[str] = None
    initrd: Optional[str] = None
    cmdline: Optional[str] = None

    dtb: Optional[str] = None

    @classmethod
    def _gen_envvar_name(cls, suffix, fields):
        # Check that all the fields are strings
        for field in fields:
            if not isinstance(field, str):
                return None

        if len(fields) > 0:
            fields_str = "_".join([f.upper() for f in fields if f]) + "_"
        else:
            fields_str = ""

        return "BOOTS_DEFAULT_" + fields_str + suffix

    @classmethod
    def _find_best_option(cls, machine, bootloader, platform, buildarch, suffix):
        # Generate all the possible names for environment variables
        prioritized_combos = [
            (bootloader, buildarch, platform),
            (buildarch, platform),
            (bootloader, buildarch),
            (buildarch, ),
            (bootloader, ),
            tuple()
        ]
        prioritized_opts = [cls._gen_envvar_name(suffix, fields) for fields in prioritized_combos]

        # Iterate through the options, in priority order, ignoring invalid entries
        for option in [o for o in prioritized_opts if o]:
            if opt := os.environ.get(option):
                return opt

    @classmethod
    def defaults(cls, machine=None, bootloader=None, platform=None, buildarch=None):
        # Possible values:
        # * bootloader: ipxe, uboot
        # * platforms: efi, pcbios, uboot
        # * buildarch: i386, x86_64, arm32, arm64

        kwargs = {"machine": machine, "bootloader": bootloader, "platform": platform, "buildarch": buildarch}
        kernel = cls._find_best_option(**kwargs, suffix="KERNEL")
        if not kernel:
            raise ValueError("No default kernel found (specified using BOOTS_DEFAULT_*KERNEL)")

        initrd = cls._find_best_option(**kwargs, suffix="INITRD")
        if not initrd:
            raise ValueError("No default initramfs found (specified using BOOTS_DEFAULT_*INITRD)")

        cmdline = cls._find_best_option(**kwargs, suffix="CMDLINE")
        if not cmdline:
            raise ValueError("No default kernel command line found (specified using BOOTS_DEFAULT_*CMDLINE)")

        # NOTE: We voluntarily do not want to use a default DTB as we want to replace this mechanism in the future
        # with something a little more flexible

        return cls(kernel=kernel, initrd=initrd, cmdline=cmdline)

    def fixup_missing_fields_with_defaults(self, machine=None, bootloader=None, platform=None, buildarch=None):
        kwargs = {"machine": machine, "bootloader": bootloader, "platform": platform, "buildarch": buildarch}

        if not self.kernel:
            self.kernel = self._find_best_option(**kwargs, suffix="KERNEL")
            if not self.kernel:
                raise ValueError("No default kernel found (specified using BOOTS_DEFAULT_*KERNEL)")

        if not self.initrd:
            self.initrd = self._find_best_option(**kwargs, suffix="INITRD")
            if not self.initrd:
                raise ValueError("No default initramfs found (specified using BOOTS_DEFAULT_*INITRD)")


class BootService:
    def __init__(self, mars,
                 private_interface=None,
                 config_paths=DEFAULT_CONFIG_PATHS):
        self.mars = mars
        self.private_interface = private_interface or config.PRIVATE_INTERFACE
        self.config_paths = config_paths

        # Do not start the servers
        if config.BOOTS_DISABLE_SERVERS:
            self.dhcpd = self.tftpd = self.fastbootd = None
            return

        # Download the iPXE binaries and store them where the DUTs can download them
        provision_ipxe_dut_clients(tftp_dir=config_paths['TFTP_DIR'])

        self.dhcpd = BootsDHCPD(self, "DHCP Server", self.private_interface)
        self.dhcpd.dns_servers = [self.dhcpd.ip]
        self.dhcpd.start()

        self.tftpd = BootsTFTPD(self, "TFTP Server", self.config_paths["TFTP_DIR"], self.private_interface)
        self.tftpd.start()

        self.fastbootd = BootsFastbootd(self, "Fastbootd Server")
        self.fastbootd.start()

    @classmethod
    def _platform_cmdline(cls, platform=None):
        return "initrd=initrd" if platform in ["efi"] else ""

    @classmethod
    def _gen_ipxe_boot_script(cls, bootconfig, platform=None):
        platform_cmdline = cls._platform_cmdline(platform=platform)

        cmdline = bootconfig.cmdline if bootconfig.cmdline is not None else ""
        cmdline = cmdline.replace(";", "${semicolon:string}")
        cmdline = cmdline.replace("&", "${ampersand:string}")
        cmdline = cmdline.replace("|", "${pipe:string}")

        return f"""#!ipxe

set semicolon:hex 3b
set ampersand:hex 26
set pipe:hex 7C

echo

echo Downloading the kernel
kernel {bootconfig.kernel} {platform_cmdline} {cmdline}

echo Downloading the initrd
initrd --name initrd {bootconfig.initrd}

echo Booting!
boot
"""

    def get_boot_config(self, machine=None, platform=None, buildarch=None, bootloader="ipxe"):
        bootconfig = None

        if machine is not None:
            bootconfig = machine.boot_config_query(platform=platform, buildarch=buildarch, bootloader=bootloader)

        if bootconfig is None:
            bootconfig = BootConfig.defaults(machine=machine, bootloader=bootloader,
                                             platform=platform, buildarch=buildarch)

        return bootconfig

    def ipxe_boot_script(self, machine=None, platform=None, buildarch=None):
        bootconfig = self.get_boot_config(machine, platform, buildarch)
        return self._gen_ipxe_boot_script(bootconfig, platform=platform)

    @classmethod
    def _gen_uboot_boot_script(cls, bootconfig, platform=None):
        platform_cmdline = cls._platform_cmdline(platform=platform)
        cmdline = bootconfig.cmdline if bootconfig.cmdline is not None else ""
        kernel = "minio" + urlparse(bootconfig.kernel).path
        initrd = "minio" + urlparse(bootconfig.initrd).path
        fdt = ("minio" + urlparse(bootconfig.dtb).path) if bootconfig.dtb else ""

        # TODO: consider adding EFI boot support?
        return f"""
echo Loading
setenv bootargs '{platform_cmdline} {cmdline}'
tftp ${{kernel_addr_r}} {kernel}
if test -n "{fdt}"
then
    tftp ${{fdt_addr}} {fdt}
elif test -n "${{fdt_file}}"
then
    tftp ${{fdt_addr}} {fdt}
else
    setenv fdt_addr ${{fdtcontroladdr}}
fi

tftp ${{ramdisk_addr_r}} {initrd}
setenv ramdisk_size ${{filesize}}
if test -z "${{kernel_comp_addr_r}}"
then
    setexpr kernel_comp_addr_r ${{ramdisk_addr_r}} + ${{ramdisk_size}}
    setenv kernel_comp_size 0x4000000 # hope for the best
fi

echo Booting!
booti ${{kernel_addr_r}} ${{ramdisk_addr_r}}:${{ramdisk_size}} ${{fdt_addr}}
bootz ${{kernel_addr_r}} ${{ramdisk_addr_r}}:${{ramdisk_size}} ${{fdt_addr}}
"""

    def uboot_boot_script(self, buildarch=None, machine_id=None):
        platform = "uboot"

        machine = self.mars.get_machine_by_id(machine_id)
        bootconfig = self.get_boot_config(machine, platform, buildarch, "uboot")

        script = bytes(self._gen_uboot_boot_script(bootconfig, platform=platform), "ascii")

        MAGIC = 0x27051956
        OS_LINUX = 5

        ARCH_INVALID = 0
        ARCH_ARM = 2
        ARCH_I386 = 3
        ARCH_ARM64 = 22
        ARCH_X86_64 = 24
        ARCH_RISCV = 26

        TYPE_SCRIPT = 6
        COMP_NONE = 0

        data = struct.pack(">II", len(script), 0) + script
        dcrc = binascii.crc32(data)

        if buildarch == CPUArch.ARM32.name:
            uboot_arch = ARCH_ARM
        elif buildarch == CPUArch.X86.name:
            uboot_arch = ARCH_I386
        elif buildarch == CPUArch.ARM64.name:
            uboot_arch = ARCH_ARM64
        elif buildarch == CPUArch.X86_64.name:
            uboot_arch = ARCH_X86_64
        elif buildarch == CPUArch.RISCV32.name or buildarch == CPUArch.RISCV64.name:
            uboot_arch = ARCH_RISCV
        else:
            uboot_arch = ARCH_INVALID

        hdr = struct.pack(">7I4b32x",
                          MAGIC, 0, int(time.time()), len(data), 0, 0, dcrc,
                          OS_LINUX, uboot_arch, TYPE_SCRIPT, COMP_NONE)
        hcrc = binascii.crc32(hdr)
        return hdr[0:4] + struct.pack(">I", hcrc) + hdr[8:] + data



from server.pdu import PDUState
from server.pdu.drivers.dummy import DummyPDU


def test_driver_DummyPDU():
    ports = ['P1', 'P2', 'P3']
    pdu = DummyPDU("MyPDU", {"ports": ports})

    assert [p.label for p in pdu.ports] == ports
    assert pdu.get_port_state(0) == PDUState.ON
    pdu.set_port_state(0, PDUState.OFF)
    assert pdu.get_port_state(0) == PDUState.OFF



from datetime import timedelta
from enum import IntEnum, auto
from threading import Thread, Event, Lock

import traceback

from . import PDU, logger


class AsyncPDUState(IntEnum):
    CREATED = auto()
    INITIALIZING = auto()
    ERROR = auto()
    STOPPED = auto()
    OK = auto()


class AsyncPDU(Thread):
    def __init__(self, model_name, pdu_name, config, reserved_port_ids=[], polling_interval=timedelta(seconds=1)):
        super().__init__(name=f'AsyncPDU-{pdu_name}')

        self.model_name = model_name
        self.pdu_name = pdu_name
        self.config = config
        self.reserved_port_ids = set(reserved_port_ids)
        self.polling_interval = polling_interval

        self.driver = None
        self.ports = []
        self.error = None
        self.port_polling_failure_count = 0

        self.started_event = Event()
        self.stop_event = Event()

    def __eq__(self, other):
        for field in ["model_name", "pdu_name", "config", "reserved_port_ids"]:
            if getattr(self, field) != getattr(other, field):
                return False
        return True

    def __hash__(self):
        return hash(id(self))

    def get_port_by_id(self, port_id, timeout=None):
        if not self.started_event.is_set():
            logger.info("Waiting for the PDU initialization to complete!")

        self.started_event.wait(timeout)

        for port in self.ports:
            if str(port.port_id) == str(port_id):
                return port

    @property
    def state(self):
        if self.native_id is None:
            return AsyncPDUState.CREATED

        if not self.is_alive():
            return AsyncPDUState.STOPPED

        if self.error:
            return AsyncPDUState.ERROR

        if not self.started_event.is_set():
            return AsyncPDUState.INITIALIZING

        return AsyncPDUState.OK

    def stop(self, wait=True):
        self.stop_event.set()

        if wait:
            self.join()

    def run(self):
        while not self.stop_event.is_set():
            self.started_event.clear()

            try:
                self.driver = PDU.create(model_name=self.model_name, pdu_name=self.pdu_name,
                                         config=self.config, reserved_port_ids=self.reserved_port_ids)
                self.ports = self.driver.ports
            except Exception as e:
                # We failed to initialize the driver, retry after a small delay
                traceback.print_exc()
                self.error = str(e)
                self.stop_event.wait(15)
                continue

            # We successfully instantiated the driver
            self.port_polling_failure_count = 0
            self.error = None
            self.started_event.set()

            while not self.stop_event.is_set():
                try:
                    self.ports = self.driver.ports
                    for port in self.ports:
                        # Update the port
                        port.state

                    # Reset the error counter
                    self.port_polling_failure_count = 0
                except Exception:  # pragma: nocover
                    # If we had 10 consecutive polling errors, try reinitializing the driver
                    self.port_polling_failure_count += 1
                    if self.port_polling_failure_count == 10:
                        break

                    traceback.print_exc()

                self.stop_event.wait(self.polling_interval.total_seconds())


class PDUDaemon:
    def __init__(self):
        self.pdus_lck = Lock()
        self.pdus = dict()

    def stop(self, wait=True):
        with self.pdus_lck:
            while len(self.pdus) > 0:
                _, pdu = self.pdus.popitem()
                pdu.stop(wait=wait)

    def get_or_create(self, *args, update_if_existing=False, **kwargs):
        new_pdu = AsyncPDU(*args, **kwargs)

        with self.pdus_lck:
            cur_pdu = self.pdus.get(new_pdu.pdu_name)

            if cur_pdu:
                if not update_if_existing:
                    # We have found a PDU which has the right name, and we don't care
                    # about updating its configuration, so let's just return it
                    return cur_pdu
                elif cur_pdu == new_pdu:
                    # We have found a PDU which has the right name, we want to update,
                    # but the config has not changed, so let's return the existing pdu
                    # after making sure that we use the expected polling interval
                    cur_pdu.polling_interval = new_pdu.polling_interval
                    return cur_pdu
                else:
                    # We have found a PDU which has the right name, but the
                    # configuration has changed so we need to delete the existing one
                    # and re-create it
                    cur_pdu.stop(wait=False)
                    del self.pdus[cur_pdu.pdu_name]

            # Start polling on the new_pdu, and add it to the list of registered PDUs
            new_pdu.start()
            self.pdus[new_pdu.pdu_name] = new_pdu

            return new_pdu

    def unregister_pdu(self, name):
        with self.pdus_lck:
            if pdu := self.pdus.pop(name, None):
                pdu.stop()



from collections import namedtuple
from pathlib import Path
from unittest.mock import patch, MagicMock, call

import pytest

from pydantic import ValidationError

from server.pdu import PDUState, PDUPort
from server.pdu.drivers.usbhub import USBHubDevice, USBHubPDU, USBHubDeviceMatchConfig, USBHubPDUConfig


DEFAULT_LOCATION = "usbaddress"


def create_usbhubdevice(tmpdir, location=DEFAULT_LOCATION, controllable_ports={1, 2, 3, 4}, **files):
    base = Path(tmpdir) / location
    base.mkdir(parents=True)

    # Create all the ports
    for port_id in controllable_ports:
        port_path = Path(base / f"{location}:1.0" / f"{location}-port{port_id}" / "disable")
        port_path.parent.mkdir(parents=True, exist_ok=True)
        port_path.write_text("0\n")

    default_files = {
        "maxchild": "4",
        "serial": None,
        "bDeviceClass": "09",
        "speed": "10000",
        "idVendor": "2109",
        "idProduct": "2817",
        "devpath": location,
    }
    for filename, default in default_files.items():
        if value := files.get(filename, default):
            (base / filename).write_text(value + "\n")

    return USBHubDevice(base)


def test_USBHubDevice__not_a_hub(tmpdir):
    with pytest.raises(ValueError) as excinfo:
        create_usbhubdevice(tmpdir, bDeviceClass="08")
    assert f"The USB device at location '{tmpdir / DEFAULT_LOCATION}' is not a USB hub" in str(excinfo)


def test_USBHubDevice__more_than_one_path_to_port_id(tmpdir):
    hub = create_usbhubdevice(tmpdir)
    hub.base_path = MagicMock(glob=MagicMock(return_value=["/path1", "/path2"]))

    with pytest.raises(ValueError) as excinfo:
        hub.port_path(1)
    assert f"Found more than one candidate port for port_id 1 in the hub at location '{hub.base_path}'" in str(excinfo)


def test_USBHubDevice__attributes(tmpdir):
    files = {
        "maxchild": "6",
        "serial": "serial",
        "bDeviceClass": "09",
        "speed": "5000",
        "idVendor": "8086",
        "idProduct": "4242",
        "devpath": DEFAULT_LOCATION,
    }

    hub = create_usbhubdevice(tmpdir, **files)

    for attr, value in files.items():
        if attr in ["maxchild", "speed"]:
            value = int(value)
        elif attr in ["bDeviceClass", "idVendor", "idProduct"]:
            value = int(value, 16)
        assert getattr(hub, attr) == value

    assert hub.controller_path == Path(tmpdir).parent


def test_USBHubDevice__set_port_state(tmpdir):
    hub = create_usbhubdevice(tmpdir)

    with pytest.raises(ValueError) as excinfo:
        hub.set_port_state(1, PDUState.REBOOT)
    assert "Unsupported state REBOOT" in str(excinfo)

    for port, state, expected_value in [(1, PDUState.ON, "0\n"), (2, PDUState.OFF, "1\n")]:
        with patch.object(hub, 'port_path') as port_path_mock:
            hub.set_port_state(port, state)
            port_path_mock.assert_called_once_with(port)
            port_path_mock.return_value.write_text.assert_called_once_with(expected_value)


def test_USBHubDevice__get_port_state(tmpdir):
    hub = create_usbhubdevice(tmpdir)

    for port, file_value, expected_state in [(1, "0\n", PDUState.ON), (2, "1\n", PDUState.OFF)]:
        with patch.object(hub, 'port_path') as port_path_mock:
            port_path_mock.return_value.read_text.return_value = file_value
            assert hub.get_port_state(port) == expected_state
            port_path_mock.assert_called_once_with(port)


def test_USBHubDeviceMatchConfig__invalid_devpath():
    for value in ["yoooo", "1.2.3."]:
        with pytest.raises(ValidationError) as excinfo:
            USBHubDeviceMatchConfig(devpath=value, idVendor=1, idProduct=2)
        assert "The devpath is not a dot-separated list of integer" in str(excinfo._excinfo)


def test_USBHubDeviceMatchConfig__matches():
    fields = dict(devpath="1.2.3", idVendor=42, idProduct=1)

    matcher = USBHubDeviceMatchConfig(**fields)
    assert matcher.matches(MagicMock(**fields))

    matcher = USBHubDeviceMatchConfig(**fields, maxchild=10)
    assert not matcher.matches(MagicMock(**fields))


def test_USBHubPDUConfig__invalid_config():
    with pytest.raises(ValidationError) as excinfo:
        USBHubPDUConfig()

    assert "Neither a `serial` nor the tuple (`controller`, `devices`) was found in the config" in str(excinfo._excinfo)


@patch("server.pdu.drivers.usbhub.USBHubDevice")
def test_USBHubPDUConfig__controller_and_devices(hub_mock, tmpdir):
    USBHubDeviceMock = namedtuple('USBHubDeviceMock', ['devpath'])

    # Check that we enforce that at least one device is set
    with pytest.raises(ValidationError) as excinfo:
        USBHubPDUConfig(controller=str(tmpdir), devices=[])
    assert "At least one device should be set" in str(excinfo._excinfo)

    # Create a list of device matchers that will respectively match any hub
    # that has a devpath == /devpath1 and devpath == /devpath2. We then make
    # sure that the mocked USBHubDevice returns a USBHubDeviceMock where only
    # the devpath is set unless the path is /devpath2 in which case we want to
    # raise an exception to simulate having an invalid hub
    dev_matchs = [
        USBHubDeviceMatchConfig(devpath="1.2.3", idVendor=0x123, idProduct=0x456),
        USBHubDeviceMatchConfig(devpath="2.3.4", idVendor=0x789, idProduct=0x123),
    ]
    dev_matchs[0].matches = MagicMock(side_effect=lambda hub: hub.devpath == "/devpath1")
    dev_matchs[1].matches = MagicMock(side_effect=lambda hub: hub.devpath == "/devpath3")
    hub_mock.side_effect = lambda path: USBHubDeviceMock(path) if path != "/devpath2" else exec("raise ValueError()")

    cfg = USBHubPDUConfig(controller=str(tmpdir), devices=dev_matchs)

    # No devices found in controller
    with pytest.raises(ValueError) as excinfo:
        cfg.hubs
    assert f"Could not find a USB device matching {dev_matchs[0]}" in str(excinfo._excinfo)

    # Right amount of devices found
    with patch.object(cfg, "controller") as controller_mock:
        controller_mock.glob.return_value = ["/devpath1", "/devpath2", "/devpath3"]

        hubs = cfg.hubs

        assert len(hubs) == len(dev_matchs)
        assert hubs[0].devpath == "/devpath1"
        assert hubs[1].devpath == "/devpath3"

        assert controller_mock.glob.call_count == len(dev_matchs)
        for dev_match in dev_matchs:
            controller_mock.glob.assert_any_call(f"usb*/*-{dev_match.devpath}")

    # More than one device returned
    with patch.object(cfg, "controller") as controller_mock:
        controller_mock.glob.return_value = ["/devpath1", "/devpath2", "/devpath3", "/devpath3"]
        with pytest.raises(ValueError) as excinfo:
            cfg.hubs
        assert f"Found more than one USB device match {dev_matchs[1]}" in str(excinfo._excinfo)


@patch("server.pdu.drivers.usbhub.USBHubDevice")
def test_USBHubPDUConfig__serial(hub_mock):
    serial = "serial1"
    parent1 = "/parent1"
    parent2 = "/parent2"

    cfg = USBHubPDUConfig(serial=f" {serial} \n")

    # Make sure that we strip the serial passed as a parameter
    assert cfg.serial == serial

    # Check the error message if no serial devices were found
    with patch.object(cfg, "USB_DEVICES_PATH") as path_mock:
        with pytest.raises(ValueError) as excinfo:
            cfg.hubs

        assert f"No USB Hubs with serial '{serial}' found" in str(excinfo._excinfo)
        path_mock.glob.assert_called_once_with("*/serial")
        hub_mock.assert_not_called()

    with patch.object(cfg, "USB_DEVICES_PATH") as path_mock:
        path_mock.glob.return_value = [
            MagicMock(read_text=MagicMock(return_value="serial0\n")),
            MagicMock(read_text=MagicMock(return_value=f"{serial}")),
            MagicMock(read_text=MagicMock(return_value=f" {serial} \n")),
            MagicMock(read_text=MagicMock(return_value="serial2")),
            MagicMock(read_text=MagicMock(return_value="serial3")),
        ]

        # Assign a parent to the paths of the devices with the right serial
        # NOTE: This could not be done directly inline because MagicMock was
        # interpreting the `parent` parameter as a MagicMock parenthood...
        path_mock.glob.return_value[1].parent = parent1
        path_mock.glob.return_value[2].parent = parent2

        assert len(cfg.hubs) == 2

        path_mock.glob.assert_called_once_with("*/serial")
        assert hub_mock.call_count == 2
        assert hub_mock.call_args_list[0] == call(parent1)
        assert hub_mock.call_args_list[1] == call(parent2)


@patch("server.pdu.drivers.usbhub.USBHubPDUConfig")
def test_USBHubPDU(usbhubcfg_mock):
    config = {"my": "config"}

    # No devices found
    usbhubcfg_mock.return_value.hubs = []
    with pytest.raises(ValueError) as excinfo:
        USBHubPDU("TestPDU", config)
    usbhubcfg_mock.assert_called_once_with(**config)
    assert "No associated hub devices found" in str(excinfo._excinfo)

    # Different controllers
    usbhubcfg_mock.return_value.hubs = [
        MagicMock(controller_path="/controller1", ports=[1, 2, 3], speed=480),
        MagicMock(controller_path="/controller2", ports=[1, 2, 3], speed=5000)
    ]
    with pytest.raises(ValueError) as excinfo:
        USBHubPDU("TestPDU", config)
    assert "Not all hubs are connected to the same USB controller" in str(excinfo._excinfo)

    # Different list of ports
    usbhubcfg_mock.return_value.hubs = [
        MagicMock(controller_path="/controller1", ports=[1, 2, 3], speed=480),
        MagicMock(controller_path="/controller1", ports=[1, 2, 4], speed=5000)
    ]
    with pytest.raises(ValueError) as excinfo:
        USBHubPDU("TestPDU", config)
    assert "Not all hubs agree on the list of ports" in str(excinfo._excinfo)

    # Same speeds
    usbhubcfg_mock.return_value.hubs = [
        MagicMock(controller_path="/controller1", ports=[1, 2, 3], speed=480),
        MagicMock(controller_path="/controller1", ports=[1, 2, 3], speed=480)
    ]
    with pytest.raises(ValueError) as excinfo:
        USBHubPDU("TestPDU", config)
    assert "Some hubs unexpectedly share the same speed" in str(excinfo._excinfo)

    # Valid configuration
    usbhubcfg_mock.return_value.hubs = [
        MagicMock(controller_path="/controller1", ports=[1, 2, 3], speed=480),
        MagicMock(controller_path="/controller1", ports=[1, 2, 3], speed=5000)
    ]
    pdu = USBHubPDU("TestPDU", config)
    assert pdu.associated_hubs == usbhubcfg_mock.return_value.hubs
    assert pdu.ports == [PDUPort(pdu, "1"), PDUPort(pdu, "2"), PDUPort(pdu, "3")]

    # Set port state
    pdu.set_port_state(1, PDUState.ON)
    for hub in usbhubcfg_mock.return_value.hubs:
        hub.set_port_state.assert_called_once_with(1, PDUState.ON)

    # Get port state -- All agree
    pdu.associated_hubs = [
        MagicMock(get_port_state=MagicMock(return_value=PDUState.ON)),
        MagicMock(get_port_state=MagicMock(return_value=PDUState.ON)),
    ]
    assert pdu.get_port_state(1) == PDUState.ON

    # Get port state -- Disagreement
    pdu.associated_hubs.append(MagicMock(get_port_state=MagicMock(return_value=PDUState.OFF)))
    assert pdu.get_port_state(1) == PDUState.UNKNOWN
